@@ -1,4 +1,4 @@
-# Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)
+# Copyright (C) 2008-2012 Association of Universities for Research in Astronomy (AURA)
 
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are met:
@@ -205,23 +205,21 @@ The order of the polynomial in the `SIP`_ ``BP_i_j`` array
 """
 
 cd = """
-``double array[2][2]``
+``double array[naxis][naxis]``
 
 The ``CDi_ja`` linear transformation matrix.
 
-For historical compatibility, two alternate specifications of the
-``CDi_ja`` and ``CROTAia`` keywords are supported.  Although these may
-not formally co-exist with ``PCi_ja``, the approach here is simply to
+For historical compatibility, three alternate specifications of the
+linear transforations are available in wcslib.  The canonical
+``PCi_ja`` with ``CDELTia``, and the deprecated ``CDi_ja`` and
+``CROTAia`` keywords.  Although the deprecated versions may not
+formally co-exist with ``PCi_ja``, the approach here is simply to
 ignore them if given in conjunction with ``PCi_ja``.
 
 `~pywcs.Wcsprm.has_pc`, `~pywcs.Wcsprm.has_cd` and
 `~pywcs.Wcsprm.has_crota` can be used to determine which of these
 alternatives are present in the header.
 
-``CDi_ja`` and ``CROTAia`` keywords, if found, are to be stored in the
-`~pywcs.Wcsprm.cd` and `~pywcs.Wcsprm.crota` arrays which are
-dimensioned similarly to `~pywcs.Wcsprm.pc` and `~pywcs.Wcsprm.cdelt`.
-
 These alternate specifications of the linear transformation matrix are
 translated immediately to ``PCi_ja`` by `~pywcs.Wcsprm.set` and are
 nowhere visible to the lower-level routines.  In particular,
@@ -330,9 +328,24 @@ crota = """
 
 ``CROTAia`` keyvalues for each coordinate axis.
 
-``CROTAia`` is an alternate specification of the linear transformation
-matrix, maintained for historical compatibility.  See `cd` for the
-current method.
+For historical compatibility, three alternate specifications of the
+linear transforations are available in wcslib.  The canonical
+``PCi_ja`` with ``CDELTia``, and the deprecated ``CDi_ja`` and
+``CROTAia`` keywords.  Although the deprecated versions may not
+formally co-exist with ``PCi_ja``, the approach here is simply to
+ignore them if given in conjunction with ``PCi_ja``.
+
+`~pywcs.Wcsprm.has_pc`, `~pywcs.Wcsprm.has_cd` and
+`~pywcs.Wcsprm.has_crota` can be used to determine which of these
+alternatives are present in the header.
+
+These alternate specifications of the linear transformation matrix are
+translated immediately to ``PCi_ja`` by `~pywcs.Wcsprm.set` and are
+nowhere visible to the lower-level routines.  In particular,
+`~pywcs.Wcsprm.set` resets `~pywcs.Wcsprm.cdelt` to unity if
+``CDi_ja`` is present (and no ``PCi_ja``).  If no ``CROTAia`` is
+associated with the latitude axis, `set` reverts to a unity ``PCi_ja``
+matrix.
 """
 
 crpix = """
@@ -415,10 +428,10 @@ the ``CRVALia``, ``CDELTia`` and ``CDi_ja`` keywords.
 
 As ``CUNITia`` is an optional header keyword, `~pywcs.Wcsprm.cunit`
 may be left blank but otherwise is expected to contain a standard
-units specification as defined by WCS Paper I.  Utility function
-`wcsutrn`, (not currently wrapped for Python) is available to
-translate commonly used non-standard units specifications but this
-must be done as a separate step before invoking `~pywcs.Wcsprm.set`.
+units specification as defined by WCS Paper I.
+`~pywcs.Wcsprm.unitfix` is available to translate commonly used
+non-standard units specifications but this must be done as a separate
+step before invoking `~pywcs.Wcsprm.set`.
 
 For celestial axes, if `~pywcs.Wcsprm.cunit` is not blank,
 `~pywcs.Wcsprm.set` uses `wcsunits` to parse it and scale
@@ -663,6 +676,30 @@ get_offset(*x, y*) -> (*x, y*)
 Returns the offset from the distortion table for pixel point (*x, y*).
 """
 
+get_cdelt = """
+get_cdelt() -> double array[naxis]
+
+Coordinate increments (``CDELTia``) for each coord axis.
+
+Returns the ``CDELT`` offsets in read-only form.  Unlike the
+`~pywcs.Wcsprm.cdelt` property, this works even when the header specifies
+the linear transformation matrix in one of the deprecated ``CDi_ja``
+or ``CROTAia`` forms.  This is useful when you want access to the
+linear transformation matrix, but don't care how it was specified in
+the header.
+"""
+
+get_pc = """
+get_pc() -> double array[naxis][naxis]
+
+Returns the ``PC`` matrix in read-only form.  Unlike the
+`~pywcs.Wcsprm.pc` property, this works even when the header specifies
+the linear transformation matrix in one of the deprecated ``CDi_ja``
+or ``CROTAia`` forms.  This is useful when you want access to the
+linear transformation matrix, but don't care how it was specified in
+the header.
+"""
+
 get_ps = """
 get_ps() -> list of tuples
 
@@ -780,6 +817,13 @@ Alias for `~pywcs.Wcsprm.has_pc`.  Maintained for backward
 compatibility.
 """
 
+have = """
+The name of the unit being converted from.
+
+This value always uses standard unit names, even if the
+`UnitConverter` was initialized with a non-standard unit name.
+"""
+
 i = """
 ``int`` (read-only)
 
@@ -1185,19 +1229,31 @@ Returns an array of focal plane coordinates.
 """ % __.ORIGIN()
 
 pc = """
-``double array[2][2]``
+``double array[naxis][naxis]``
 
 The ``PCi_ja`` (pixel coordinate) transformation matrix.  The order is::
 
   [[PC1_1, PC1_2],
    [PC2_1, PC2_2]]
 
-For historical compatibility, two alternate specifications of the
-``CDi_ja`` and ``CROTAia`` keywords are supported.
+For historical compatibility, three alternate specifications of the
+linear transforations are available in wcslib.  The canonical
+``PCi_ja`` with ``CDELTia``, and the deprecated ``CDi_ja`` and
+``CROTAia`` keywords.  Although the deprecated versions may not
+formally co-exist with ``PCi_ja``, the approach here is simply to
+ignore them if given in conjunction with ``PCi_ja``.
 
-`~pywcs.Wcsprm.has_pci_ja`, `~pywcs.Wcsprm.has_cdi_ja` and
-`~pywcs.Wcsprm.has_crotaia` can be used to determine which of these
+`~pywcs.Wcsprm.has_pc`, `~pywcs.Wcsprm.has_cd` and
+`~pywcs.Wcsprm.has_crota` can be used to determine which of these
 alternatives are present in the header.
+
+These alternate specifications of the linear transformation matrix are
+translated immediately to ``PCi_ja`` by `~pywcs.Wcsprm.set` and are
+nowhere visible to the lower-level routines.  In particular,
+`~pywcs.Wcsprm.set` resets `~pywcs.Wcsprm.cdelt` to unity if
+``CDi_ja`` is present (and no ``PCi_ja``).  If no ``CROTAia`` is
+associated with the latitude axis, `set` reverts to a unity ``PCi_ja``
+matrix.
 """
 
 phi0 = """
@@ -1449,7 +1505,7 @@ sequence of tuples of the form (*i*, *m*, *value*):
 
     - *m*: int.  Parameter number, as in ``PVi_ma``, (i.e. 0-relative)
 
-    - *value*: string.  Parameter value.
+    - *value*: float.  Parameter value.
 
 .. seealso::
 
@@ -1927,6 +1983,13 @@ An undefined value is represented by NaN.
    `~pywcs.Wcsprm.specsys`, `~pywcs.Wcsprm.ssysobs`
 """
 
+want = """
+The name of the unit being converted to.
+
+This value always uses standard unit names, even if the
+`UnitConverter` was initialized with a non-standard unit name.
+"""
+
 wcs = """
 A `~pywcs.Wcsprm` object to perform the basic `wcslib`_ WCS
 tranformation.
