@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)
+Copyright (C) 2008-2012 Association of Universities for Research in Astronomy (AURA)
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
@@ -118,7 +118,7 @@ PyWcs_dealloc(
   int ignored;
   ignored = PyWcs_clear(self);
   pipeline_free(&self->x);
-  self->ob_type->tp_free((PyObject*)self);
+  Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
 /*@null@*/ static PyObject *
@@ -268,6 +268,7 @@ PyWcs_all_pix2sky(
   }
 
   /* Make the call */
+  Py_BEGIN_ALLOW_THREADS
   preoffset_array(pixcrd, origin);
   wcsprm_python2c(self->x.wcs);
   status = pipeline_all_pixel2world(&self->x,
@@ -277,6 +278,7 @@ PyWcs_all_pix2sky(
                                     (double*)PyArray_DATA(world));
   wcsprm_c2python(self->x.wcs);
   unoffset_array(pixcrd, origin);
+  Py_END_ALLOW_THREADS
   /* unoffset_array(world, origin); */
 
  exit:
@@ -291,14 +293,11 @@ PyWcs_all_pix2sky(
     return NULL;
   } else {
     Py_DECREF(world);
-    if (status > 0 && status < WCS_ERRMSG_MAX) {
-      PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
-      return NULL;
-    } else if (status == -1) {
+    if (status == -1) {
+      /* exception already set */
       return NULL;
     } else {
-      PyErr_SetString(PyExc_RuntimeError,
-                      "Unknown error occurred.  Something is seriously wrong.");
+      wcserr_to_python_exc(self->x.err);
       return NULL;
     }
   }
@@ -344,6 +343,7 @@ PyWcs_p4_pix2foc(
     goto exit;
   }
 
+  Py_BEGIN_ALLOW_THREADS
   preoffset_array(pixcrd, origin);
   status = p4_pix2foc(2, (void *)self->x.cpdis,
                       (unsigned int)PyArray_DIM(pixcrd, 0),
@@ -351,6 +351,7 @@ PyWcs_p4_pix2foc(
                       (double*)PyArray_DATA(foccrd));
   unoffset_array(pixcrd, origin);
   unoffset_array(foccrd, origin);
+  Py_END_ALLOW_THREADS
 
  exit:
 
@@ -360,14 +361,11 @@ PyWcs_p4_pix2foc(
     return (PyObject*)foccrd;
   } else {
     Py_XDECREF(foccrd);
-    if (status > 0 && status < WCS_ERRMSG_MAX) {
-      PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
-      return NULL;
-    } else if (status == -1) {
+    if (status == -1) {
+      /* Exception already set */
       return NULL;
     } else {
-      PyErr_SetString(PyExc_RuntimeError,
-                      "Unknown error occurred.  Something is seriously wrong.");
+      PyErr_SetString(PyExc_MemoryError, "NULL pointer passed");
       return NULL;
     }
   }
@@ -413,6 +411,7 @@ PyWcs_det2im(
     goto exit;
   }
 
+  Py_BEGIN_ALLOW_THREADS
   preoffset_array(detcrd, origin);
   status = p4_pix2foc(2, (void *)self->x.det2im,
                       (unsigned int)PyArray_DIM(detcrd, 0),
@@ -420,6 +419,7 @@ PyWcs_det2im(
                       (double*)PyArray_DATA(imcrd));
   unoffset_array(detcrd, origin);
   unoffset_array(imcrd, origin);
+  Py_END_ALLOW_THREADS
 
  exit:
 
@@ -429,14 +429,11 @@ PyWcs_det2im(
     return (PyObject*)imcrd;
   } else {
     Py_XDECREF(imcrd);
-    if (status > 0 && status < WCS_ERRMSG_MAX) {
-      PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
-      return NULL;
-    } else if (status == -1) {
+    if (status == -1) {
+      /* Exception already set */
       return NULL;
     } else {
-      PyErr_SetString(PyExc_RuntimeError,
-                      "Unknown error occurred.  Something is seriously wrong.");
+      PyErr_SetString(PyExc_MemoryError, "NULL pointer passed");
       return NULL;
     }
   }
@@ -476,6 +473,7 @@ PyWcs_pix2foc(
     goto _exit;
   }
 
+  Py_BEGIN_ALLOW_THREADS
   preoffset_array(pixcrd, origin);
   status = pipeline_pix2foc(&self->x,
                             (unsigned int)PyArray_DIM(pixcrd, 0),
@@ -484,6 +482,7 @@ PyWcs_pix2foc(
                             (double*)PyArray_DATA(foccrd));
   unoffset_array(pixcrd, origin);
   unoffset_array(foccrd, origin);
+  Py_END_ALLOW_THREADS
 
  _exit:
 
@@ -493,14 +492,11 @@ PyWcs_pix2foc(
     return (PyObject*)foccrd;
   } else {
     Py_XDECREF(foccrd);
-    if (status > 0 && status < WCS_ERRMSG_MAX) {
-      PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
-      return NULL;
-    } else if (status == -1) {
+    if (status == -1) {
+      /* Exception already set */
       return NULL;
     } else {
-      PyErr_SetString(PyExc_RuntimeError,
-                      "Unknown error occurred.  Something is seriously wrong.");
+      wcserr_to_python_exc(self->x.err);
       return NULL;
     }
   }
@@ -871,14 +867,18 @@ static PyMethodDef PyWcs_methods[] = {
 };
 
 static PyMethodDef module_methods[] = {
-    {"_sanity_check", (PyCFunction)_sanity_check, METH_NOARGS, ""},
-    {"find_all_wcs", (PyCFunction)PyWcsprm_find_all_wcs, METH_VARARGS|METH_KEYWORDS, doc_find_all_wcs},
-    {NULL}  /* Sentinel */
+  {"_sanity_check", (PyCFunction)_sanity_check, METH_NOARGS, ""},
+  {"find_all_wcs", (PyCFunction)PyWcsprm_find_all_wcs, METH_VARARGS|METH_KEYWORDS, doc_find_all_wcs},
+  {NULL}  /* Sentinel */
 };
 
 static PyTypeObject PyWcsType = {
+  #if PY3K
+  PyVarObject_HEAD_INIT(NULL, 0)
+  #else
   PyObject_HEAD_INIT(NULL)
   0,                            /*ob_size*/
+  #endif
   "pywcs._Wcs",                 /*tp_name*/
   sizeof(PyWcs),                /*tp_basicsize*/
   0,                            /*tp_itemsize*/
@@ -940,14 +940,7 @@ struct module_state {
 #endif
 };
 
-#if PY_MAJOR_VERSION >= 3
-    #define GETSTATE(m) ((struct module_state*)PyModule_GetState(m))
-#else
-    #define GETSTATE(m) (&_state)
-    static struct module_state _state;
-#endif
-
-#if PY_MAJOR_VERSION >= 3
+#if PY3K
     static struct PyModuleDef moduledef = {
         PyModuleDef_HEAD_INIT,
         "_pywcs",
@@ -962,63 +955,63 @@ struct module_state {
 
     #define INITERROR return NULL
 
-    PyObject *
+    PyMODINIT_FUNC
     PyInit__pywcs(void)
 
 #else
     #define INITERROR return
 
-    void
+    PyMODINIT_FUNC
     init_pywcs(void)
 #endif
 
 {
-    PyObject* m;
+  PyObject* m;
 
-    wcs_errexc[0] = NULL;                         /* Success */
-    wcs_errexc[1] = &PyExc_MemoryError;           /* Null wcsprm pointer passed */
-    wcs_errexc[2] = &PyExc_MemoryError;           /* Memory allocation failed */
-    wcs_errexc[3] = &WcsExc_SingularMatrix;       /* Linear transformation matrix is singular */
-    wcs_errexc[4] = &WcsExc_InconsistentAxisTypes; /* Inconsistent or unrecognized coordinate axis types */
-    wcs_errexc[5] = &PyExc_ValueError;            /* Invalid parameter value */
-    wcs_errexc[6] = &WcsExc_InvalidTransform;     /* Invalid coordinate transformation parameters */
-    wcs_errexc[7] = &WcsExc_InvalidTransform;     /* Ill-conditioned coordinate transformation parameters */
-    wcs_errexc[8] = &WcsExc_InvalidCoordinate;    /* One or more of the pixel coordinates were invalid, */
-    /* as indicated by the stat vector */
-    wcs_errexc[9] = &WcsExc_InvalidCoordinate;    /* One or more of the world coordinates were invalid, */
-                                        /* as indicated by the stat vector */
-    wcs_errexc[10] = &WcsExc_InvalidCoordinate;    /* Invalid world coordinate */
-    wcs_errexc[11] = &WcsExc_NoSolution;           /* no solution found in the specified interval */
-    wcs_errexc[12] = &WcsExc_InvalidSubimageSpecification; /* Invalid subimage specification (no spectral axis) */
-    wcs_errexc[13] = &WcsExc_NonseparableSubimageCoordinateSystem; /* Non-separable subimage coordinate system */
+  wcs_errexc[0] = NULL;                         /* Success */
+  wcs_errexc[1] = &PyExc_MemoryError;           /* Null wcsprm pointer passed */
+  wcs_errexc[2] = &PyExc_MemoryError;           /* Memory allocation failed */
+  wcs_errexc[3] = &WcsExc_SingularMatrix;       /* Linear transformation matrix is singular */
+  wcs_errexc[4] = &WcsExc_InconsistentAxisTypes; /* Inconsistent or unrecognized coordinate axis types */
+  wcs_errexc[5] = &PyExc_ValueError;            /* Invalid parameter value */
+  wcs_errexc[6] = &WcsExc_InvalidTransform;     /* Invalid coordinate transformation parameters */
+  wcs_errexc[7] = &WcsExc_InvalidTransform;     /* Ill-conditioned coordinate transformation parameters */
+  wcs_errexc[8] = &WcsExc_InvalidCoordinate;    /* One or more of the pixel coordinates were invalid, */
+  /* as indicated by the stat vector */
+  wcs_errexc[9] = &WcsExc_InvalidCoordinate;    /* One or more of the world coordinates were invalid, */
+  /* as indicated by the stat vector */
+  wcs_errexc[10] = &WcsExc_InvalidCoordinate;    /* Invalid world coordinate */
+  wcs_errexc[11] = &WcsExc_NoSolution;           /* no solution found in the specified interval */
+  wcs_errexc[12] = &WcsExc_InvalidSubimageSpecification; /* Invalid subimage specification (no spectral axis) */
+  wcs_errexc[13] = &WcsExc_NonseparableSubimageCoordinateSystem; /* Non-separable subimage coordinate system */
 
-#if PY_MAJOR_VERSION >= 3
-    m = PyModule_Create(&moduledef);
+#if PY3K
+  m = PyModule_Create(&moduledef);
 #else
-    m = Py_InitModule3("_pywcs", module_methods, NULL);
+  m = Py_InitModule3("_pywcs", module_methods, NULL);
 #endif
 
-    if (m == NULL)
-        INITERROR;
+  if (m == NULL)
+    INITERROR;
 
-    import_array();
-    fill_docstrings();
+  import_array();
+  fill_docstrings();
 
-    if (_setup_api(m)                 ||
-        _setup_str_list_proxy_type(m) ||
-        _setup_wcsprm_type(m)         ||
-        _setup_tabprm_type(m)         ||
-        _setup_units_type(m)          ||
-        /* _setup_wtbarr_type(m)         || */
-        _setup_distortion_type(m)     ||
-        _setup_sip_type(m)            ||
-        _setup_pywcs_type(m)          ||
-        _define_exceptions(m)) {
-        Py_DECREF(m);
-        INITERROR;
-    }
+  if (_setup_api(m)                 ||
+      _setup_str_list_proxy_type(m) ||
+      _setup_wcsprm_type(m)         ||
+      _setup_tabprm_type(m)         ||
+      _setup_units_type(m)          ||
+      /* _setup_wtbarr_type(m)         || */
+      _setup_distortion_type(m)     ||
+      _setup_sip_type(m)            ||
+      _setup_pywcs_type(m)          ||
+      _define_exceptions(m)) {
+    Py_DECREF(m);
+    INITERROR;
+  }
 
-#if PY_MAJOR_VERSION >= 3
-    return module;
+#if PY3K
+  return m;
 #endif
 }
