@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2008 Association of Universities for Research in Astronomy (AURA)
+Copyright (C) 2008-2012 Association of Universities for Research in Astronomy (AURA)
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
@@ -39,6 +39,7 @@ DAMAGE.
 #include "wcslib_wrap.h"
 #include "wcslib_tabprm_wrap.h"
 #include "wcslib_wtbarr_wrap.h"
+#include "wcslib_units_wrap.h"
 #include <structmember.h> /* from Python */
 
 #include <wcs.h>
@@ -46,9 +47,9 @@ DAMAGE.
 #include <wcshdr.h>
 #include <wcsmath.h>
 #include <wcsprintf.h>
+#include <wcsunits.h>
 
 #include "isnan.h"
-#include "str_list_proxy.h"
 #include "distortion.h"
 
 /*
@@ -88,7 +89,7 @@ is_valid_alt_key(
  */
 
 static int
-PyWcsprm_cset(PyWcsprm* self);
+PyWcsprm_cset(PyWcsprm* self, const int convert);
 
 static inline void
 note_change(PyWcsprm* self) {
@@ -101,7 +102,7 @@ PyWcsprm_dealloc(
 
   int ignored;
   ignored = wcsfree(&self->x);
-  self->ob_type->tp_free((PyObject*)self);
+  Py_TYPE(self)->tp_free((PyObject*)self);
 }
 
 static PyWcsprm*
@@ -199,20 +200,24 @@ PyWcsprm_init(
     if (status != 0) {
       PyErr_SetString(
           PyExc_MemoryError,
-          "Could not initialize wcsprm object");
+          self->x.err->msg);
       return -1;
     }
 
     self->x.alt[0] = key[0];
 
-    if (PyWcsprm_cset(self)) {
+    if (PyWcsprm_cset(self, 0)) {
       return -1;
     }
     wcsprm_c2python(&self->x);
 
     return 0;
   } else { /* header != NULL */
+    #if PY3K
+    if (PyBytes_AsStringAndSize(header_obj, &header, &header_length)) {
+    #else
     if (PyString_AsStringAndSize(header_obj, &header, &header_length)) {
+    #endif
       return -1;
     }
 
@@ -221,7 +226,11 @@ PyWcsprm_init(
     } else if (relax_obj == NULL || relax_obj == Py_False) {
       relax = WCSHDR_none;
     } else {
+      #if PY3K
+      relax = PyLong_AsLong(relax_obj);
+      #else
       relax = PyInt_AsLong(relax_obj);
+      #endif
       if (relax == -1) {
         PyErr_SetString(
             PyExc_ValueError,
@@ -331,14 +340,11 @@ PyWcsprm_init(
       ignored_int = wcsvfree(&nwcs, &wcs);
       PyErr_SetString(
           PyExc_MemoryError,
-          "Could not initialize wcsprm object");
+          self->x.err->msg);
       return -1;
     }
 
     note_change(self);
-    if (PyWcsprm_cset(self)) {
-      return -1;
-    }
     wcsprm_c2python(&self->x);
     ignored_int = wcsvfree(&nwcs, &wcs);
     return 0;
@@ -362,7 +368,7 @@ PyWcsprm_copy(
   wcsprm_c2python(&self->x);
 
   if (status == 0) {
-    if (PyWcsprm_cset(copy)) {
+    if (PyWcsprm_cset(copy, 0)) {
       Py_XDECREF(copy);
       return NULL;
     }
@@ -370,15 +376,8 @@ PyWcsprm_copy(
     return (PyObject*)copy;
   } else {
     Py_XDECREF(copy);
-    if (status > 0 && status < WCS_ERRMSG_MAX) {
-      PyErr_SetString(*wcs_errexc[status], wcscopy_errmsg[status]);
-      return NULL;
-    } else {
-      PyErr_SetString(
-          PyExc_RuntimeError,
-          "Unknown error occurred.  Something is seriously wrong.");
-      return NULL;
-    }
+    wcs_to_python_exc(&(self->x));
+    return NULL;
   }
 }
 
@@ -412,7 +411,11 @@ PyWcsprm_find_all_wcs(
     return NULL;
   }
 
+  #if PY3K
+  if (PyBytes_AsStringAndSize(header_obj, &header, &header_length)) {
+  #else
   if (PyString_AsStringAndSize(header_obj, &header, &header_length)) {
+  #endif
     return NULL;
   }
 
@@ -429,7 +432,11 @@ PyWcsprm_find_all_wcs(
   } else if (relax_obj == NULL || relax_obj == Py_False) {
     relax = WCSHDR_none;
   } else {
+    #if PY3K
+    relax = PyLong_AsLong(relax_obj);
+    #else
     relax = PyInt_AsLong(relax_obj);
+    #endif
     if (relax == -1) {
       PyErr_SetString(
           PyExc_ValueError,
@@ -438,6 +445,7 @@ PyWcsprm_find_all_wcs(
     }
   }
 
+  Py_BEGIN_ALLOW_THREADS
   if (keysel < 0) {
     status = wcspih(
         header,
@@ -459,6 +467,7 @@ PyWcsprm_find_all_wcs(
         &nwcs,
         &wcs);
   }
+  Py_END_ALLOW_THREADS
 
   if (status != 0) {
     PyErr_SetString(
@@ -492,7 +501,7 @@ PyWcsprm_find_all_wcs(
     }
 
     subresult->x.flag = 0;
-    if (PyWcsprm_cset(subresult)) {
+    if (PyWcsprm_cset(subresult, 0)) {
       Py_DECREF(subresult);
       Py_DECREF(result);
       ignored = wcsvfree(&nwcs, &wcs);
@@ -516,14 +525,13 @@ PyWcsprm_celfix(
   wcsprm_c2python(&self->x);
 
   if (status == -1 || status == 0) {
+    #if PY3K
+    return PyLong_FromLong((long)status);
+    #else
     return PyInt_FromLong((long)status);
-  } else if (status > 0 && status < WCSFIX_ERRMSG_MAX) {
-    PyErr_SetString(PyExc_ValueError, wcsfix_errmsg[status]);
-    return NULL;
+    #endif
   } else {
-    PyErr_SetString(
-        PyExc_RuntimeError,
-        "Unknown error occurred.  Something is seriously wrong.");
+    wcserr_fix_to_python_exc(self->x.err);
     return NULL;
   }
 }
@@ -572,14 +580,13 @@ PyWcsprm_cylfix(
   Py_XDECREF(naxis_array);
 
   if (status == -1 || status == 0) {
+    #if PY3K
+    return PyLong_FromLong((long)status);
+    #else
     return PyInt_FromLong((long)status);
-  } else if (status > 0 && status < WCSFIX_ERRMSG_MAX) {
-    PyErr_SetString(PyExc_ValueError, wcsfix_errmsg[status]);
-    return NULL;
+    #endif
   } else {
-    PyErr_SetString(
-        PyExc_RuntimeError,
-        "Unknown error occurred.  Something is seriously wrong.");
+    wcserr_fix_to_python_exc(self->x.err);
     return NULL;
   }
 }
@@ -590,17 +597,18 @@ PyWcsprm_datfix(
 
   int status = 0;
 
+  wcsprm_python2c(&self->x);
   status = datfix(&self->x);
+  wcsprm_c2python(&self->x);
 
   if (status == -1 || status == 0) {
+    #if PY3K
+    return PyLong_FromLong((long)status);
+    #else
     return PyInt_FromLong((long)status);
-  } else if (status > 0 && status < WCSFIX_ERRMSG_MAX) {
-    PyErr_SetString(PyExc_ValueError, wcsfix_errmsg[status]);
-    return NULL;
+    #endif
   } else {
-    PyErr_SetString(
-        PyExc_RuntimeError,
-        "Unknown error occurred.  Something is seriously wrong.");
+    wcserr_fix_to_python_exc(self->x.err);
     return NULL;
   }
 }
@@ -617,11 +625,13 @@ PyWcsprm_fix(
   PyArrayObject* naxis_array     = NULL;
   int*           naxis           = NULL;
   int            stat[NWCSFIX];
+  struct wcserr  err[NWCSFIX];
   int            status          = 0;
   PyObject*      subresult;
   PyObject*      result;
   int            i               = 0;
   int            msg_index       = 0;
+  const char*    message;
   PyObject*      ignored         = NULL;
 
   struct message_map_entry {
@@ -668,8 +678,9 @@ PyWcsprm_fix(
     naxis = (int*)PyArray_DATA(naxis_array);
   }
 
+  /* TODO: Use wcsfixi */
   wcsprm_python2c(&self->x);
-  status = wcsfix(ctrl, naxis, &self->x, stat);
+  status = wcsfixi(ctrl, naxis, &self->x, stat, err);
   wcsprm_c2python(&self->x);
 
   /* We're done with this already, so deref now so we don't have to remember
@@ -683,21 +694,73 @@ PyWcsprm_fix(
 
   for (i = 0; i < 5; ++i) {
     msg_index = stat[message_map[i].index];
-    if (msg_index >= 0 && msg_index < 11) {
-      subresult = PyString_FromString(wcsfix_errmsg[msg_index]);
-      if (subresult == NULL ||
-          PyDict_SetItemString(result, message_map[i].name, subresult)) {
-        Py_XDECREF(subresult);
-        Py_XDECREF(result);
-        return NULL;
-      }
-      Py_XDECREF(subresult);
+    if (msg_index > 0 && msg_index < 11) {
+      message = err[message_map[i].index].msg;
+    } else if (msg_index == 0) {
+      message = "Success";
+    } else {
+      message = "No change";
     }
+    #if PY3K
+    subresult = PyUnicode_FromString(message);
+    #else
+    subresult = PyString_FromString(message);
+    #endif
+    if (subresult == NULL ||
+        PyDict_SetItemString(result, message_map[i].name, subresult)) {
+      Py_XDECREF(subresult);
+      Py_XDECREF(result);
+      return NULL;
+    }
+    Py_XDECREF(subresult);
   }
 
   return result;
 }
 
+/*@null@*/ static PyObject*
+PyWcsprm_get_cdelt_func(
+    PyWcsprm* self,
+    /*@unused@*/ PyObject* args,
+    /*@unused@*/ PyObject* kwds) {
+
+  Py_ssize_t naxis = 0;
+
+  if (is_null(self->x.cdelt)) {
+    return NULL;
+  }
+
+  if (PyWcsprm_cset(self, 1)) {
+    return NULL;
+  }
+
+  naxis = self->x.naxis;
+
+  return get_double_array_readonly("cdelt", self->x.cdelt, 1, &naxis, (PyObject*)self);
+}
+
+/*@null@*/ static PyObject*
+PyWcsprm_get_pc_func(
+    PyWcsprm* self,
+    /*@unused@*/ PyObject* args,
+    /*@unused@*/ PyObject* kwds) {
+
+  npy_intp dims[2];
+
+  if (is_null(self->x.pc)) {
+    return NULL;
+  }
+
+  if (PyWcsprm_cset(self, 1)) {
+    return NULL;
+  }
+
+  dims[0] = self->x.naxis;
+  dims[1] = self->x.naxis;
+
+  return get_double_array_readonly("pc", self->x.pc, 2, dims, (PyObject*)self);
+}
+
 /*@null@*/ static PyObject*
 PyWcsprm_get_ps(
     PyWcsprm* self,
@@ -767,7 +830,7 @@ static PyObject*
 PyWcsprm_is_unity(
     PyWcsprm* self) {
 
-  if (PyWcsprm_cset(self)) {
+  if (PyWcsprm_cset(self, 1)) {
     return NULL;
   }
 
@@ -848,7 +911,7 @@ PyWcsprm_mix(
     goto exit;
   }
 
-  if (mixpix < 1 || mixpix > 2) {
+  if (mixpix < 1 || mixpix > self->x.naxis) {
     PyErr_SetString(
         PyExc_ValueError,
         "Argument 1 (mixpix) must specify a pixel coordinate "
@@ -871,25 +934,23 @@ PyWcsprm_mix(
   phi = (PyArrayObject*)PyArray_SimpleNew
     (1, &naxis, PyArray_DOUBLE);
   if (phi == NULL) {
-    status = 2;
     goto exit;
   }
 
   theta = (PyArrayObject*)PyArray_SimpleNew
     (1, &naxis, PyArray_DOUBLE);
   if (theta == NULL) {
-    status = 2;
     goto exit;
   }
 
   imgcrd = (PyArrayObject*)PyArray_SimpleNew
     (1, &naxis, PyArray_DOUBLE);
   if (imgcrd == NULL) {
-    status = 2;
     goto exit;
   }
 
   /* Convert pixel coordinates to 1-based */
+  Py_BEGIN_ALLOW_THREADS
   preoffset_array(pixcrd, origin);
   wcsprm_python2c(&self->x);
   status = wcsmix(
@@ -907,6 +968,7 @@ PyWcsprm_mix(
   wcsprm_c2python(&self->x);
   unoffset_array(pixcrd, origin);
   unoffset_array(imgcrd, origin);
+  Py_END_ALLOW_THREADS
 
   if (status == 0) {
     result = PyDict_New();
@@ -915,7 +977,7 @@ PyWcsprm_mix(
         PyDict_SetItemString(result, "phi", (PyObject*)phi) ||
         PyDict_SetItemString(result, "theta", (PyObject*)theta) ||
         PyDict_SetItemString(result, "world", (PyObject*)world)) {
-      status = 2;
+      goto exit;
     }
   }
 
@@ -933,13 +995,8 @@ PyWcsprm_mix(
     if (status == -1) {
       /* The error message has already been set */
       return NULL;
-    } else if (status > 0 && status < WCS_ERRMSG_MAX) {
-      PyErr_SetString(*wcs_errexc[status], wcsmix_errmsg[status]);
-      return NULL;
     } else {
-      PyErr_SetString(
-          PyExc_RuntimeError,
-          "Unknown error occurred.  Something is seriously wrong.");
+      wcs_to_python_exc(&(self->x));
       return NULL;
     }
   }
@@ -952,6 +1009,8 @@ PyWcsprm_p2s(
     PyObject* kwds) {
 
   int            naxis      = 2;
+  int            ncoord     = 0;
+  int            nelem      = 0;
   PyObject*      pixcrd_obj = NULL;
   int            origin     = 1;
   PyArrayObject* pixcrd     = NULL;
@@ -992,45 +1051,43 @@ PyWcsprm_p2s(
   imgcrd = (PyArrayObject*)PyArray_SimpleNew(
       2, PyArray_DIMS(pixcrd), PyArray_DOUBLE);
   if (imgcrd == NULL) {
-    status = 2;
     goto exit;
   }
 
   phi = (PyArrayObject*)PyArray_SimpleNew(
       1, PyArray_DIMS(pixcrd), PyArray_DOUBLE);
   if (phi == NULL) {
-    status = 2;
     goto exit;
   }
 
   theta = (PyArrayObject*)PyArray_SimpleNew(
       1, PyArray_DIMS(pixcrd), PyArray_DOUBLE);
   if (theta == NULL) {
-    status = 2;
     goto exit;
   }
 
   world = (PyArrayObject*)PyArray_SimpleNew(
       2, PyArray_DIMS(pixcrd), PyArray_DOUBLE);
   if (world == NULL) {
-    status = 2;
     goto exit;
   }
 
   stat = (PyArrayObject*)PyArray_SimpleNew(
       1, PyArray_DIMS(pixcrd), PyArray_INT);
   if (stat == NULL) {
-    status = 2;
     goto exit;
   }
 
-  preoffset_array(pixcrd, origin);
   /* Make the call */
+  Py_BEGIN_ALLOW_THREADS
+  ncoord = PyArray_DIM(pixcrd, 0);
+  nelem = PyArray_DIM(pixcrd, 1);
+  preoffset_array(pixcrd, origin);
   wcsprm_python2c(&self->x);
   status = wcsp2s(
       &self->x,
-      (int)PyArray_DIM(pixcrd, 0),
-      (int)PyArray_DIM(pixcrd, 1),
+      ncoord,
+      nelem,
       (double*)PyArray_DATA(pixcrd),
       (double*)PyArray_DATA(imgcrd),
       (double*)PyArray_DATA(phi),
@@ -1041,6 +1098,17 @@ PyWcsprm_p2s(
   unoffset_array(pixcrd, origin);
   /* unoffset_array(world, origin); */
   unoffset_array(imgcrd, origin);
+  if (status == 8) {
+    set_invalid_to_nan(
+        ncoord, nelem, (double*)PyArray_DATA(imgcrd), (int*)PyArray_DATA(stat));
+    set_invalid_to_nan(
+        ncoord, 1, (double*)PyArray_DATA(phi), (int*)PyArray_DATA(stat));
+    set_invalid_to_nan(
+        ncoord, 1, (double*)PyArray_DATA(theta), (int*)PyArray_DATA(stat));
+    set_invalid_to_nan(
+        ncoord, nelem, (double*)PyArray_DATA(world), (int*)PyArray_DATA(stat));
+  }
+  Py_END_ALLOW_THREADS
 
   if (status == 0 || status == 8) {
     result = PyDict_New();
@@ -1050,7 +1118,7 @@ PyWcsprm_p2s(
         PyDict_SetItemString(result, "theta", (PyObject*)theta) ||
         PyDict_SetItemString(result, "world", (PyObject*)world) ||
         PyDict_SetItemString(result, "stat", (PyObject*)stat)) {
-      status = 2;
+      goto exit;
     }
   }
 
@@ -1066,15 +1134,11 @@ PyWcsprm_p2s(
     return result;
   } else {
     Py_XDECREF(result);
-    if (status > 0 && status < WCS_ERRMSG_MAX) {
-      PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
-      return NULL;
-    } else if (status == -1) {
+    if (status == -1) {
+      /* Exception already set */
       return NULL;
     } else {
-      PyErr_SetString(
-          PyExc_RuntimeError,
-          "Unknown error occurred.  Something is seriously wrong.");
+      wcs_to_python_exc(&(self->x));
       return NULL;
     }
   }
@@ -1087,6 +1151,8 @@ PyWcsprm_s2p(
     PyObject* kwds) {
 
   int            naxis     = 2;
+  int            ncoord    = 0;
+  int            nelem     = 0;
   PyObject*      world_obj = NULL;
   int            origin    = 1;
   PyArrayObject* world     = NULL;
@@ -1096,7 +1162,7 @@ PyWcsprm_s2p(
   PyArrayObject* pixcrd    = NULL;
   PyArrayObject* stat      = NULL;
   PyObject*      result    = NULL;
-  int            status    = 0;
+  int            status    = -1;
   const char*    keywords[] = {
     "world", "origin", NULL };
 
@@ -1128,45 +1194,43 @@ PyWcsprm_s2p(
   phi = (PyArrayObject*)PyArray_SimpleNew(
       1, PyArray_DIMS(world), PyArray_DOUBLE);
   if (phi == NULL) {
-    status = 2;
     goto exit;
   }
 
   theta = (PyArrayObject*)PyArray_SimpleNew(
       1, PyArray_DIMS(world), PyArray_DOUBLE);
   if (phi == NULL) {
-    status = 2;
     goto exit;
   }
 
   imgcrd = (PyArrayObject*)PyArray_SimpleNew(
       2, PyArray_DIMS(world), PyArray_DOUBLE);
   if (theta == NULL) {
-    status = 2;
     goto exit;
   }
 
   pixcrd = (PyArrayObject*)PyArray_SimpleNew(
       2, PyArray_DIMS(world), PyArray_DOUBLE);
   if (pixcrd == NULL) {
-    status = 2;
     goto exit;
   }
 
   stat = (PyArrayObject*)PyArray_SimpleNew(
       1, PyArray_DIMS(world), PyArray_INT);
   if (stat == NULL) {
-    status = 2;
     goto exit;
   }
 
   /* Make the call */
+  Py_BEGIN_ALLOW_THREADS
+  ncoord = PyArray_DIM(world, 0);
+  nelem = PyArray_DIM(world, 1);
   /* preoffset_array(world, origin); */
   wcsprm_python2c(&self->x);
   status = wcss2p(
       &self->x,
-      (int)PyArray_DIM(world, 0),
-      (int)PyArray_DIM(world, 1),
+      ncoord,
+      nelem,
       (double*)PyArray_DATA(world),
       (double*)PyArray_DATA(phi),
       (double*)PyArray_DATA(theta),
@@ -1177,6 +1241,17 @@ PyWcsprm_s2p(
   /* unoffset_array(world, origin); */
   unoffset_array(pixcrd, origin);
   unoffset_array(imgcrd, origin);
+  if (status == 8) {
+    set_invalid_to_nan(
+        ncoord, 1, (double*)PyArray_DATA(phi), (int*)PyArray_DATA(stat));
+    set_invalid_to_nan(
+        ncoord, 1, (double*)PyArray_DATA(theta), (int*)PyArray_DATA(stat));
+    set_invalid_to_nan(
+        ncoord, nelem, (double*)PyArray_DATA(imgcrd), (int*)PyArray_DATA(stat));
+    set_invalid_to_nan(
+        ncoord, nelem, (double*)PyArray_DATA(pixcrd), (int*)PyArray_DATA(stat));
+  }
+  Py_END_ALLOW_THREADS
 
   if (status == 0 || status == 9) {
     result = PyDict_New();
@@ -1186,7 +1261,7 @@ PyWcsprm_s2p(
         PyDict_SetItemString(result, "imgcrd", (PyObject*)imgcrd) ||
         PyDict_SetItemString(result, "pixcrd", (PyObject*)pixcrd) ||
         PyDict_SetItemString(result, "stat", (PyObject*)stat)) {
-      status = 2;
+      goto exit;
     }
   }
 
@@ -1202,13 +1277,11 @@ PyWcsprm_s2p(
     return result;
   } else {
     Py_XDECREF(result);
-    if (status > 0 && status < WCS_ERRMSG_MAX) {
-      PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
+    if (status == -1) {
+      /* Exception already set */
       return NULL;
     } else {
-      PyErr_SetString(
-          PyExc_RuntimeError,
-          "Unknown error occurred.  Something is seriously wrong.");
+      wcs_to_python_exc(&(self->x));
       return NULL;
     }
   }
@@ -1216,24 +1289,20 @@ PyWcsprm_s2p(
 
 static int
 PyWcsprm_cset(
-    PyWcsprm* self) {
+    PyWcsprm* self,
+    const int convert) {
 
   int status = 0;
 
-  wcsprm_python2c(&self->x);
+  if (convert) wcsprm_python2c(&self->x);
   status = wcsset(&self->x);
-  wcsprm_c2python(&self->x);
+  if (convert) wcsprm_c2python(&self->x);
 
   if (status == 0) {
     return 0;
-  } else if (status > 0 && status < WCS_ERRMSG_MAX) {
-    PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
-    return -1;
   } else {
-    PyErr_SetString(
-        PyExc_RuntimeError,
-        "Unknown error occurred.  Something is seriously wrong.");
-    return -1;
+    wcs_to_python_exc(&(self->x));
+    return 1;
   }
 }
 
@@ -1241,7 +1310,7 @@ PyWcsprm_cset(
 PyWcsprm_set(
     PyWcsprm* self) {
 
-  if (PyWcsprm_cset(self)) {
+  if (PyWcsprm_cset(self, 1)) {
     return NULL;
   }
 
@@ -1260,8 +1329,10 @@ PyWcsprm_set_ps(
   }
 
   if (set_pscards("ps", arg, &self->x.ps, &self->x.nps, &self->x.npsmax)) {
+    self->x.m_ps = self->x.ps;
     return NULL;
   }
+  self->x.m_ps = self->x.ps;
 
   note_change(self);
 
@@ -1280,8 +1351,10 @@ PyWcsprm_set_pv(
   }
 
   if (set_pvcards("pv", arg, &self->x.pv, &self->x.npv, &self->x.npvmax)) {
+    self->x.m_pv = self->x.pv;
     return NULL;
   }
+  self->x.m_pv = self->x.pv;
 
   note_change(self);
 
@@ -1298,15 +1371,19 @@ PyWcsprm_print_contents(
 
   int ignored;
 
-  if (PyWcsprm_cset(self)) {
-    return NULL;
-  }
+  /* This is not thread-safe, but since we're holding onto the GIL,
+     we can assume we won't have thread conflicts */
+  wcsprintf_set(NULL);
 
   wcsprm_python2c(&self->x);
+  if (PyWcsprm_cset(self, 0)) {
+    wcsprm_c2python(&self->x);
+    return NULL;
+  }
   ignored = wcsprt(&self->x);
   wcsprm_c2python(&self->x);
 
-  printf(wcsprintf_buf());
+  printf("%s", wcsprintf_buf());
 
   Py_INCREF(Py_None);
   return Py_None;
@@ -1323,14 +1400,13 @@ PyWcsprm_spcfix(
   wcsprm_c2python(&self->x);
 
   if (status == -1 || status == 0) {
+    #if PY3K
+    return PyLong_FromLong((long)status);
+    #else
     return PyInt_FromLong((long)status);
-  } else if (status > 0 && status < WCSFIX_ERRMSG_MAX) {
-    PyErr_SetString(PyExc_ValueError, wcsfix_errmsg[status]);
-    return NULL;
+    #endif
   } else {
-    PyErr_SetString(
-        PyExc_RuntimeError,
-        "Unknown error occurred.  Something is seriously wrong.");
+    wcserr_fix_to_python_exc(self->x.err);
     return NULL;
   }
 }
@@ -1368,13 +1444,8 @@ PyWcsprm_sptr(
   if (status == 0) {
     Py_INCREF(Py_None);
     return Py_None;
-  } else if (status > 0 && status < WCS_ERRMSG_MAX) {
-    PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
-    return NULL;
   } else {
-    PyErr_SetString(
-        PyExc_RuntimeError,
-        "Unknown error occurred.  Something is seriously wrong.");
+    wcs_to_python_exc(&(self->x));
     return NULL;
   }
 }
@@ -1383,15 +1454,23 @@ PyWcsprm_sptr(
 PyWcsprm___str__(
     PyWcsprm* self) {
 
-  if (PyWcsprm_cset(self)) {
-    return NULL;
-  }
+  /* This is not thread-safe, but since we're holding onto the GIL,
+     we can assume we won't have thread conflicts */
+  wcsprintf_set(NULL);
 
   wcsprm_python2c(&self->x);
+  if (PyWcsprm_cset(self, 0)) {
+    wcsprm_c2python(&self->x);
+    return NULL;
+  }
   wcsprt(&self->x);
   wcsprm_c2python(&self->x);
 
+  #if PY3K
+  return PyUnicode_FromString(wcsprintf_buf());
+  #else
   return PyString_FromString(wcsprintf_buf());
+  #endif
 }
 
 /*@null@*/ static PyObject*
@@ -1405,6 +1484,9 @@ PyWcsprm_sub(
   PyObject*  py_axes        = NULL;
   PyWcsprm*  py_dest_wcs    = NULL;
   PyObject*  element        = NULL;
+  #if PY3K
+  PyObject*  element_utf8   = NULL;
+  #endif
   char*      element_str    = NULL;
   int        element_val    = 0;
   int        nsub           = 0;
@@ -1440,8 +1522,19 @@ PyWcsprm_sub(
         goto exit;
       }
 
+      #if PY3K
+      if (PyUnicode_Check(element)) {
+        element_utf8 = PyUnicode_AsUTF8String(element);
+        if (element_utf8 == NULL) {
+          goto exit;
+        }
+        element_str = PyBytes_AsString(element_utf8);
+        Py_DECREF(element_utf8); element_utf8 = NULL;
+      #else
       if (PyString_Check(element)) {
+        /* Doesn't return NULL, because we already known it's a string */
         element_str = PyString_AsString(element);
+      #endif
         if (strncmp(element_str, "longitude", 10) == 0) {
           element_val = WCSSUB_LONGITUDE;
         } else if (strncmp(element_str, "latitude", 9) == 0) {
@@ -1460,8 +1553,13 @@ PyWcsprm_sub(
             "string values for axis sequence must be one of 'latitude', 'longitude', 'cubeface', 'spectral', 'stokes', or 'celestial'");
           goto exit;
         }
+      #if PY3K
+      } else if (PyLong_Check(element)) {
+        tmp = (Py_ssize_t)PyLong_AsSsize_t(element);
+      #else
       } else if (PyInt_Check(element)) {
         tmp = (Py_ssize_t)PyInt_AsLong(element);
+      #endif
         if (tmp == -1 && PyErr_Occurred()) {
           goto exit;
         }
@@ -1478,8 +1576,13 @@ PyWcsprm_sub(
       Py_DECREF(element);
       element = NULL;
     }
+  #if PY3K
+  } else if (PyLong_Check(py_axes)) {
+    tmp = (Py_ssize_t)PyLong_AsSsize_t(py_axes);
+  #else
   } else if (PyInt_Check(py_axes)) {
     tmp = (Py_ssize_t)PyInt_AsLong(py_axes);
+  #endif
     if (tmp == -1 && PyErr_Occurred()) {
       goto exit;
     }
@@ -1515,7 +1618,7 @@ PyWcsprm_sub(
   wcsprm_python2c(&self->x);
   status = wcssub(0, &self->x, &nsub, axes, &py_dest_wcs->x);
   wcsprm_c2python(&self->x);
-  if (PyWcsprm_cset(py_dest_wcs)) {
+  if (PyWcsprm_cset(py_dest_wcs, 0)) {
     goto exit;
   }
   wcsprm_c2python(&py_dest_wcs->x);
@@ -1524,24 +1627,22 @@ PyWcsprm_sub(
     goto exit;
   }
 
-  Py_INCREF(py_dest_wcs);
-
  exit:
   free(axes);
   Py_XDECREF(element);
-  Py_XDECREF(py_dest_wcs);
+  #if PY3K
+  Py_XDECREF(element_utf8);
+  #endif
 
   if (status == 0) {
     return (PyObject*)py_dest_wcs;
-  } else if (status > 0 && status < WCS_ERRMSG_MAX) {
-    PyErr_SetString(*wcs_errexc[status], wcsp2s_errmsg[status]);
-    return NULL;
-  } else if (status < 0) {
+  } else if (status == -1) {
+    Py_XDECREF(py_dest_wcs);
+    /* Exception already set */
     return NULL;
   } else {
-    PyErr_SetString(
-        PyExc_RuntimeError,
-        "Unknown error occurred.  Something is seriously wrong.");
+    wcs_to_python_exc(&(py_dest_wcs->x));
+    Py_XDECREF(py_dest_wcs);
     return NULL;
   }
 }
@@ -1556,7 +1657,7 @@ PyWcsprm_to_header(
   int       relax        = 0;
   int       nkeyrec      = 0;
   char*     header       = NULL;
-  int       status       = 0;
+  int       status       = -1;
   PyObject* result       = NULL;
   const char* keywords[] = {"relax", NULL};
 
@@ -1571,7 +1672,11 @@ PyWcsprm_to_header(
   } else if (relax_obj == NULL || relax_obj == Py_False) {
     relax = WCSHDO_none;
   } else {
+    #if PY3K
+    relax = PyLong_AsLong(relax_obj);
+    #else
     relax = PyInt_AsLong(relax_obj);
+    #endif
     if (relax == -1) {
       PyErr_SetString(
           PyExc_ValueError,
@@ -1593,7 +1698,11 @@ PyWcsprm_to_header(
 
   /* Just return the raw header string.  PyFITS on the Python side will help
      to parse and use this information. */
+  #if PY3K
+  result = PyBytes_FromStringAndSize(header, (Py_ssize_t)nkeyrec * 80);
+  #else
   result = PyString_FromStringAndSize(header, (Py_ssize_t)nkeyrec * 80);
+  #endif
 
  exit:
   free(header);
@@ -1626,14 +1735,13 @@ PyWcsprm_unitfix(
   status = unitfix(ctrl, &self->x);
 
   if (status == -1 || status == 0) {
+    #if PY3K
+    return PyLong_FromLong((long)status);
+    #else
     return PyInt_FromLong((long)status);
-  } else if (status > 0 && status < WCSFIX_ERRMSG_MAX) {
-    PyErr_SetString(PyExc_ValueError, wcsfix_errmsg[status]);
-    return NULL;
+    #endif
   } else {
-    PyErr_SetString(
-        PyExc_RuntimeError,
-        "Unknown error occurred.  Something is seriously wrong.");
+    wcserr_fix_to_python_exc(self->x.err);
     return NULL;
   }
 }
@@ -1701,6 +1809,10 @@ PyWcsprm_get_axis_types(
     return NULL;
   }
 
+  if (PyWcsprm_cset(self, 1)) {
+    return NULL;
+  }
+
   naxis = (Py_ssize_t)self->x.naxis;
 
   return get_int_array("axis_types", self->x.types, 1, &naxis, (PyObject*)self);
@@ -1711,7 +1823,7 @@ PyWcsprm_get_cd(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  const npy_intp dims[2] = {2, 2};
+  npy_intp dims[2];
 
   if (is_null(self->x.cd)) {
     return NULL;
@@ -1722,6 +1834,9 @@ PyWcsprm_get_cd(
     return NULL;
   }
 
+  dims[0] = self->x.naxis;
+  dims[1] = self->x.naxis;
+
   return get_double_array("cd", self->x.cd, 2, dims, (PyObject*)self);
 }
 
@@ -1731,7 +1846,7 @@ PyWcsprm_set_cd(
     PyObject* value,
     /*@unused@*/ void* closure) {
 
-  const npy_intp dims[2] = {2, 2};
+  npy_intp dims[2];
 
   if (is_null(self->x.cd)) {
     return -1;
@@ -1744,6 +1859,9 @@ PyWcsprm_set_cd(
     return 0;
   }
 
+  dims[0] = self->x.naxis;
+  dims[1] = self->x.naxis;
+
   if (set_double_array("cd", value, 2, dims, self->x.cd)) {
     return -1;
   }
@@ -1755,33 +1873,7 @@ PyWcsprm_set_cd(
   return 0;
 }
 
-/*@null@*/ static PyObject*
-PyWcsprm_get_cname(
-    PyWcsprm* self,
-    /*@unused@*/ void* closure) {
-
-  if (is_null(self->x.cname)) {
-    return NULL;
-  }
-
-  return get_str_list("cname", self->x.cname, (Py_ssize_t)self->x.naxis, (PyObject*)self);
-}
-
-/*@null@*/ static int
-PyWcsprm_set_cname(
-    PyWcsprm* self,
-    PyObject* value,
-    /*@unused@*/ void* closure) {
-  if (is_null(self->x.cname)) {
-    return -1;
-  }
-
-  note_change(self);
-
-  return set_str_list("cname", value, (Py_ssize_t)self->x.naxis, 0, self->x.cname);
-}
-
-/*@null@*/ static PyObject*
+ /*@null@*/ static PyObject*
 PyWcsprm_get_cdelt(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
@@ -1829,7 +1921,7 @@ PyWcsprm_get_cel_offset(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  return PyBool_FromLong(self->x.cel.offset);
+  return get_bool("cel_offset", self->x.cel.offset);
 }
 
 static int
@@ -1843,23 +1935,31 @@ PyWcsprm_set_cel_offset(
   return set_bool("cel_offset", value, &self->x.cel.offset);
 }
 
-static PyObject*
-PyWcsprm_get_colnum(
+
+/*@null@*/ static PyObject*
+PyWcsprm_get_cname(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  return get_int("colnum", self->x.colnum);
+  if (is_null(self->x.cname)) {
+    return NULL;
+  }
+
+  return get_str_list("cname", self->x.cname, (Py_ssize_t)self->x.naxis, 68, (PyObject*)self);
 }
 
-static int
-PyWcsprm_set_colnum(
+/*@null@*/ static int
+PyWcsprm_set_cname(
     PyWcsprm* self,
     PyObject* value,
     /*@unused@*/ void* closure) {
+  if (is_null(self->x.cname)) {
+    return -1;
+  }
 
   note_change(self);
 
-  return set_int("colnum", value, &self->x.colnum);
+  return set_str_list("cname", value, (Py_ssize_t)self->x.naxis, 0, self->x.cname);
 }
 
 /*@null@*/ static PyObject*
@@ -1897,6 +1997,25 @@ PyWcsprm_set_colax(
   return set_int_array("colax", value, 1, &naxis, self->x.colax);
 }
 
+static PyObject*
+PyWcsprm_get_colnum(
+    PyWcsprm* self,
+    /*@unused@*/ void* closure) {
+
+  return get_int("colnum", self->x.colnum);
+}
+
+static int
+PyWcsprm_set_colnum(
+    PyWcsprm* self,
+    PyObject* value,
+    /*@unused@*/ void* closure) {
+
+  note_change(self);
+
+  return set_int("colnum", value, &self->x.colnum);
+}
+
 /*@null@*/ static PyObject*
 PyWcsprm_get_crder(
     PyWcsprm* self,
@@ -2098,7 +2217,7 @@ PyWcsprm_get_ctype(
     return NULL;
   }
 
-  return get_str_list("ctype", self->x.ctype, self->x.naxis, (PyObject*)self);
+  return get_str_list("ctype", self->x.ctype, self->x.naxis, 68, (PyObject*)self);
 }
 
 static int
@@ -2121,7 +2240,7 @@ PyWcsprm_get_cubeface(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  return get_bool("cubeface", self->x.cubeface);
+  return get_int("cubeface", self->x.cubeface);
 }
 
 static int
@@ -2132,7 +2251,24 @@ PyWcsprm_set_cubeface(
 
   note_change(self);
 
-  return set_bool("cubeface", value, &self->x.cubeface);
+  return set_int("cubeface", value, &self->x.cubeface);
+}
+
+static int
+unit_verify(char* val) {
+
+  int status, func;
+  double scale, units[WCSUNITS_NTYPE];
+  struct wcserr *err = NULL;
+
+  status = wcsulexe(val, &func, &scale, units, &err);
+  if (status == 0) {
+    return 1;
+  } else {
+    wcserr_units_to_python_exc(err);
+    free(err);
+    return 0;
+  }
 }
 
 /*@null@*/ static PyObject*
@@ -2144,7 +2280,9 @@ PyWcsprm_get_cunit(
     return NULL;
   }
 
-  return get_str_list("cunit", self->x.cunit, (Py_ssize_t)self->x.naxis, (PyObject*)self);
+  return get_str_list_verified(
+    "cunit", self->x.cunit, (Py_ssize_t)self->x.naxis, 68, (PyObject*)self,
+    unit_verify);
 }
 
 static int
@@ -2159,7 +2297,9 @@ PyWcsprm_set_cunit(
 
   note_change(self);
 
-  return set_str_list("cunit", value, (Py_ssize_t)self->x.naxis, 0, self->x.cunit);
+  return set_str_list_verified(
+    "cunit", value, (Py_ssize_t)self->x.naxis, 0, self->x.cunit,
+    unit_verify);
 }
 
 /*@null@*/ static PyObject*
@@ -2186,6 +2326,8 @@ PyWcsprm_set_dateavg(
 
   note_change(self);
 
+  /* TODO: Verify that this looks like a date string */
+
   return set_string("dateavg", value, self->x.dateavg, 72);
 }
 
@@ -2245,16 +2387,19 @@ PyWcsprm_get_imgpix_matrix(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  const npy_intp dims[2] = {2, 2};
+  npy_intp dims[2];
 
   if (is_null(self->x.lin.imgpix)) {
     return NULL;
   }
 
-  if (PyWcsprm_cset(self)) {
+  if (PyWcsprm_cset(self, 1)) {
     return NULL;
   }
 
+  dims[0] = self->x.naxis;
+  dims[1] = self->x.naxis;
+
   return get_double_array("imgpix_matrix", self->x.lin.imgpix, 2, dims,
                           (PyObject*)self);
 }
@@ -2264,7 +2409,7 @@ PyWcsprm_get_lat(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  if (PyWcsprm_cset(self)) {
+  if (PyWcsprm_cset(self, 1)) {
     return NULL;
   }
 
@@ -2287,6 +2432,11 @@ PyWcsprm_set_latpole(
 
   note_change(self);
 
+  if (value == NULL) {
+    self->x.latpole = 90.0;
+    return 0;
+  }
+
   return set_double("latpole", value, &self->x.latpole);
 }
 
@@ -2299,7 +2449,7 @@ PyWcsprm_get_lattyp(
     return NULL;
   }
 
-  if (PyWcsprm_cset(self)) {
+  if (PyWcsprm_cset(self, 1)) {
     return NULL;
   }
 
@@ -2311,7 +2461,7 @@ PyWcsprm_get_lng(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  if (PyWcsprm_cset(self)) {
+  if (PyWcsprm_cset(self, 1)) {
     return NULL;
   }
 
@@ -2327,7 +2477,7 @@ PyWcsprm_get_lngtyp(
     return NULL;
   }
 
-  if (PyWcsprm_cset(self)) {
+  if (PyWcsprm_cset(self, 1)) {
     return NULL;
   }
 
@@ -2350,6 +2500,11 @@ PyWcsprm_set_lonpole(
 
   note_change(self);
 
+  if (value == NULL) {
+    self->x.lonpole = (double)NPY_NAN;
+    return 0;
+  }
+
   return set_double("lonpole", value, &self->x.lonpole);
 }
 
@@ -2369,6 +2524,11 @@ PyWcsprm_set_mjdavg(
 
   note_change(self);
 
+  if (value == NULL) {
+    self->x.mjdavg = (double)NPY_NAN;
+    return 0;
+  }
+
   return set_double("mjdavg", value, &self->x.mjdavg);
 }
 
@@ -2388,6 +2548,11 @@ PyWcsprm_set_mjdobs(
 
   note_change(self);
 
+  if (value == NULL) {
+    self->x.mjdobs = (double)NPY_NAN;
+    return 0;
+  }
+
   return set_double("mjdobs", value, &self->x.mjdobs);
 }
 
@@ -2454,6 +2619,13 @@ PyWcsprm_set_obsgeo(
 
   note_change(self);
 
+  if (value == NULL) {
+    self->x.obsgeo[0] = NPY_NAN;
+    self->x.obsgeo[1] = NPY_NAN;
+    self->x.obsgeo[2] = NPY_NAN;
+    return 0;
+  }
+
   return set_double_array("obsgeo", value, 1, &size, self->x.obsgeo);
 }
 
@@ -2462,7 +2634,7 @@ PyWcsprm_get_pc(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  const npy_intp dims[2] = {2, 2};
+  npy_intp dims[2];
 
   if (is_null(self->x.pc)) {
     return NULL;
@@ -2473,6 +2645,9 @@ PyWcsprm_get_pc(
     return NULL;
   }
 
+  dims[0] = self->x.naxis;
+  dims[1] = self->x.naxis;
+
   return get_double_array("pc", self->x.pc, 2, dims, (PyObject*)self);
 }
 
@@ -2482,7 +2657,7 @@ PyWcsprm_set_pc(
     PyObject* value,
     /*@unused@*/ void* closure) {
 
-  const npy_intp dims[2] = {2, 2};
+  npy_intp dims[2];
   int i, j, naxis;
   double* pc;
 
@@ -2515,6 +2690,9 @@ PyWcsprm_set_pc(
     return 0;
   }
 
+  dims[0] = self->x.naxis;
+  dims[1] = self->x.naxis;
+
   if (set_double_array("pc", value, 2, dims, self->x.pc)) {
     return -1;
   }
@@ -2542,6 +2720,11 @@ PyWcsprm_set_phi0(
 
   note_change(self);
 
+  if (value == NULL) {
+    self->x.cel.phi0 = (double)NPY_NAN;
+    return 0;
+  }
+
   return set_double("phi0", value, &(self->x.cel.phi0));
 }
 
@@ -2550,16 +2733,19 @@ PyWcsprm_get_piximg_matrix(
     PyWcsprm* self,
     /*@unused@*/ void* closure) {
 
-  const npy_intp dims[2] = {2, 2};
+  npy_intp dims[2];
 
   if (is_null(self->x.lin.piximg)) {
     return NULL;
   }
 
-  if (PyWcsprm_cset(self)) {
+  if (PyWcsprm_cset(self, 1)) {
     return NULL;
   }
 
+  dims[0] = self->x.naxis;
+  dims[1] = self->x.naxis;
+
   return get_double_array("piximg_matrix", self->x.lin.piximg, 2, dims,
                           (PyObject*)self);
 }
@@ -2777,6 +2963,11 @@ PyWcsprm_set_theta0(
 
   note_change(self);
 
+  if (value == NULL) {
+    self->x.cel.theta0 = (double)NPY_NAN;
+    return 0;
+  }
+
   return set_double("theta0", value, &self->x.cel.theta0);
 }
 
@@ -2947,6 +3138,8 @@ static PyMethodDef PyWcsprm_methods[] = {
   {"datfix", (PyCFunction)PyWcsprm_datfix, METH_NOARGS, doc_datfix},
   {"__deepcopy__", (PyCFunction)PyWcsprm_copy, METH_O, doc_copy},
   {"fix", (PyCFunction)PyWcsprm_fix, METH_VARARGS|METH_KEYWORDS, doc_fix},
+  {"get_cdelt", (PyCFunction)PyWcsprm_get_cdelt_func, METH_NOARGS, doc_get_cdelt},
+  {"get_pc", (PyCFunction)PyWcsprm_get_pc_func, METH_NOARGS, doc_get_pc},
   {"get_ps", (PyCFunction)PyWcsprm_get_ps, METH_NOARGS, doc_get_ps},
   {"get_pv", (PyCFunction)PyWcsprm_get_pv, METH_NOARGS, doc_get_pv},
   {"has_cd", (PyCFunction)PyWcsprm_has_cdi_ja, METH_NOARGS, doc_has_cd},
@@ -2972,8 +3165,12 @@ static PyMethodDef PyWcsprm_methods[] = {
 };
 
 PyTypeObject PyWcsprmType = {
+  #if PY3K
+  PyVarObject_HEAD_INIT(NULL, 0)
+  #else
   PyObject_HEAD_INIT(NULL)
   0,                            /*ob_size*/
+  #endif
   "pywcs._Wcsprm",              /*tp_name*/
   sizeof(PyWcsprm),             /*tp_basicsize*/
   0,                            /*tp_itemsize*/
@@ -3026,6 +3223,7 @@ _setup_wcsprm_type(
   Py_INCREF(&PyWcsprmType);
 
   wcsprintf_set(NULL);
+  wcserr_enable(1);
 
   return (
     PyModule_AddObject(m, "_Wcsprm", (PyObject *)&PyWcsprmType) ||
