@@ -1,6 +1,6 @@
 /*============================================================================
 
-  WCSLIB 4.7 - an implementation of the FITS WCS standard.
+  WCSLIB 4.8 - an implementation of the FITS WCS standard.
   Copyright (C) 1995-2011, Mark Calabretta
 
   This file is part of WCSLIB.
@@ -28,13 +28,15 @@
 
   Author: Mark Calabretta, Australia Telescope National Facility
   http://www.atnf.csiro.au/~mcalabre/index.html
-  $Id: spc.c,v 4.7 2011/02/07 07:03:42 cal103 Exp $
+  $Id: spc.c,v 4.8.1.1 2011/08/15 08:07:06 cal103 Exp cal103 $
 *===========================================================================*/
 
 #include <math.h>
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 
+#include "wcserr.h"
 #include "wcsmath.h"
 #include "wcsprintf.h"
 #include "wcstrig.h"
@@ -43,25 +45,25 @@
 #include "spx.h"
 
 /* Spectral algorithm codes. */
-#define F2S 100;                /* Axis linear in frequency.          */
-#define W2S 200;                /* Axis linear in vacuum wavelengths. */
-#define A2S 300;                /* Axis linear in air wavelengths.    */
-#define V2S 400;                /* Axis linear in velocity.           */
-#define GRI 500;                /* Grism in vacuum.                   */
-#define GRA 600;                /* Grism in air.                      */
+#define F2S 100;		/* Axis linear in frequency.          */
+#define W2S 200;		/* Axis linear in vacuum wavelengths. */
+#define A2S 300;		/* Axis linear in air wavelengths.    */
+#define V2S 400;		/* Axis linear in velocity.           */
+#define GRI 500;		/* Grism in vacuum.                   */
+#define GRA 600;		/* Grism in air.                      */
 
 /* S-type spectral variables. */
-#define FREQ  0;                /* Frequency-like.                    */
-#define AFRQ  1;                /* Frequency-like.                    */
-#define ENER  2;                /* Frequency-like.                    */
-#define WAVN  3;                /* Frequency-like.                    */
-#define VRAD  4;                /* Frequency-like.                    */
-#define WAVE 10;                /* Vacuum wavelength-like.            */
-#define VOPT 11;                /* Vacuum wavelength-like.            */
-#define ZOPT 12;                /* Vacuum wavelength-like.            */
-#define AWAV 20;                /* Air wavelength-like.               */
-#define VELO 30;                /* Velocity-like.                     */
-#define BETA 31;                /* Velocity-like.                     */
+#define FREQ  0;		/* Frequency-like.                    */
+#define AFRQ  1;		/* Frequency-like.                    */
+#define ENER  2;		/* Frequency-like.                    */
+#define WAVN  3;		/* Frequency-like.                    */
+#define VRAD  4;		/* Frequency-like.                    */
+#define WAVE 10;		/* Vacuum wavelength-like.            */
+#define VOPT 11;		/* Vacuum wavelength-like.            */
+#define ZOPT 12;		/* Vacuum wavelength-like.            */
+#define AWAV 20;		/* Air wavelength-like.               */
+#define VELO 30;		/* Velocity-like.                     */
+#define BETA 31;		/* Velocity-like.                     */
 
 
 /* Map status return value to message. */
@@ -72,6 +74,9 @@ const char *spc_errmsg[] = {
   "One or more of x coordinates were invalid",
   "One or more of the spec coordinates were invalid"};
 
+/* Convenience macro for invoking wcserr_set(). */
+#define SPC_ERRMSG(status) WCSERR_SET(status), spc_errmsg[status]
+
 
 #define C 2.99792458e8
 
@@ -82,7 +87,7 @@ int spcini(struct spcprm *spc)
 {
   register int k;
 
-  if (spc == 0x0) return 1;
+  if (spc == 0x0) return SPCERR_NULL_POINTER;
 
   spc->flag = 0;
 
@@ -103,6 +108,8 @@ int spcini(struct spcprm *spc)
 
   spc->isGrism = 0;
 
+  spc->err = 0x0;
+
   spc->spxX2P = 0x0;
   spc->spxP2S = 0x0;
   spc->spxS2P = 0x0;
@@ -113,12 +120,28 @@ int spcini(struct spcprm *spc)
 
 /*--------------------------------------------------------------------------*/
 
+int spcfree(spc)
+
+struct spcprm *spc;
+
+{
+  if (spc == 0x0) return SPCERR_NULL_POINTER;
+
+  if (spc->err) free(spc->err);
+  spc->err = 0x0;
+
+  return 0;
+}
+
+/*--------------------------------------------------------------------------*/
+
 int spcprt(const struct spcprm *spc)
 
 {
-  int i;
+  char hext[32];
+  int  i;
 
-  if (spc == 0x0) return 1;
+  if (spc == 0x0) return SPCERR_NULL_POINTER;
 
   wcsprintf("       flag: %d\n", spc->flag);
   wcsprintf("       type: \"%s\"\n", spc->type);
@@ -169,10 +192,20 @@ int spcprt(const struct spcprm *spc)
   }
 
   wcsprintf("    isGrism: %d\n", spc->isGrism);
-  wcsprintf("     spxX2P: %p\n", (void *)spc->spxX2P);
-  wcsprintf("     spxP2S: %p\n", (void *)spc->spxP2S);
-  wcsprintf("     spxS2P: %p\n", (void *)spc->spxS2P);
-  wcsprintf("     spxP2X: %p\n", (void *)spc->spxP2X);
+
+  WCSPRINTF_PTR("        err: ", spc->err, "\n");
+  if (spc->err) {
+    wcserr_prt(spc->err, "");
+  }
+
+  wcsprintf("     spxX2P: %s\n",
+    wcsutil_fptr2str((int (*)())spc->spxX2P, hext));
+  wcsprintf("     spxP2S: %s\n",
+    wcsutil_fptr2str((int (*)())spc->spxP2S, hext));
+  wcsprintf("     spxS2P: %s\n",
+    wcsutil_fptr2str((int (*)())spc->spxS2P, hext));
+  wcsprintf("     spxP2X: %s\n",
+    wcsutil_fptr2str((int (*)())spc->spxP2X, hext));
 
   return 0;
 }
@@ -182,15 +215,20 @@ int spcprt(const struct spcprm *spc)
 int spcset(struct spcprm *spc)
 
 {
+  static const char *function = "spcset";
+
   char   ctype[9], ptype, xtype;
   int    restreq, status;
   double alpha, beta_r, crvalX, dn_r, dXdS, epsilon, G, m, lambda_r, n_r,
          t, restfrq, restwav, theta;
+  struct wcserr **err;
 
-  if (spc == 0x0) return 1;
+  if (spc == 0x0) return SPCERR_NULL_POINTER;
+  err = &(spc->err);
 
   if (undefined(spc->crval)) {
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Spectral crval is undefined");
   }
 
   spc->type[4] = '\0';
@@ -202,8 +240,8 @@ int spcset(struct spcprm *spc)
   sprintf(ctype, "%s-%s", spc->type, spc->code);
   restfrq = spc->restfrq;
   restwav = spc->restwav;
-  if ((status = spcspx(ctype, spc->crval, restfrq, restwav, &ptype, &xtype,
-                       &restreq, &crvalX, &dXdS))) {
+  if ((status = spcspxe(ctype, spc->crval, restfrq, restwav, &ptype, &xtype,
+                        &restreq, &crvalX, &dXdS, &(spc->err)))) {
     return status;
   }
 
@@ -468,18 +506,22 @@ int spcx2s(
   int stat[])
 
 {
+  static const char *function = "spcx2s";
+
   int statP2S, status = 0, statX2P;
   double beta;
   register int ix;
   register int *statp;
   register const double *xp;
   register double *specp;
-
+  struct wcserr **err;
 
   /* Initialize. */
-  if (spc == 0x0) return 1;
+  if (spc == 0x0) return SPCERR_NULL_POINTER;
+  err = &(spc->err);
+
   if (spc->flag == 0) {
-    if (spcset(spc)) return 2;
+    if ((status = spcset(spc))) return status;
   }
 
   /* Convert intermediate world coordinate x to X. */
@@ -505,10 +547,13 @@ int spcx2s(
   if (spc->spxX2P) {
     if ((statX2P = spc->spxX2P(spc->w[0], nx, sspec, sspec, spec, spec,
                                stat))) {
-      if (statX2P == 4) {
-        status = 3;
+      if (statX2P == SPXERR_BAD_INSPEC_COORD) {
+        status = SPCERR_BAD_X;
+      } else if (statX2P == SPXERR_BAD_SPEC_PARAMS) {
+        return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+          "Invalid spectral parameters: Frequency or wavelength is 0");
       } else {
-        return statX2P;
+        return wcserr_set(SPC_ERRMSG(statX2P));
       }
     }
   }
@@ -518,14 +563,20 @@ int spcx2s(
   if (spc->spxP2S) {
     if ((statP2S = spc->spxP2S(spc->w[0], nx, sspec, sspec, spec, spec,
                                stat))) {
-      if (statP2S == 4) {
-        status = 3;
+      if (statP2S == SPXERR_BAD_INSPEC_COORD) {
+        status = SPCERR_BAD_X;
+      } else if (statP2S == SPXERR_BAD_SPEC_PARAMS) {
+        return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+          "Invalid spectral parameters: Frequency or wavelength is 0");
       } else {
-        return statP2S;
+        return wcserr_set(SPC_ERRMSG(statP2S));
       }
     }
   }
 
+  if (status) {
+    wcserr_set(SPC_ERRMSG(status));
+  }
   return status;
 }
 
@@ -541,28 +592,35 @@ int spcs2x(
   int stat[])
 
 {
+  static const char *function = "spcs2x";
+
   int statP2X, status = 0, statS2P;
   double beta, s;
   register int ispec;
   register int *statp;
   register const double *specp;
   register double *xp;
-
+  struct wcserr **err;
 
   /* Initialize. */
-  if (spc == 0x0) return 1;
+  if (spc == 0x0) return SPCERR_NULL_POINTER;
+  err = &(spc->err);
+
   if (spc->flag == 0) {
-    if (spcset(spc)) return 2;
+    if ((status = spcset(spc))) return status;
   }
 
   /* Apply the linear step of the algorithm chain to convert the S-type */
   /* spectral variable to P-type intermediate spectral variable.        */
   if (spc->spxS2P) {
     if ((statS2P = spc->spxS2P(spc->w[0], nspec, sspec, sx, spec, x, stat))) {
-      if (statS2P == 4) {
-        status = 4;
+      if (statS2P == SPXERR_BAD_INSPEC_COORD) {
+        status = SPCERR_BAD_SPEC;
+      } else if (statS2P == SPXERR_BAD_SPEC_PARAMS) {
+        return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+          "Invalid spectral parameters: Frequency or wavelength is 0");
       } else {
-        return statS2P;
+        return wcserr_set(SPC_ERRMSG(statS2P));
       }
     }
 
@@ -582,10 +640,13 @@ int spcs2x(
   /* intermediate spectral variable to X-type spectral variable. */
   if (spc->spxP2X) {
     if ((statP2X = spc->spxP2X(spc->w[0], nspec, sx, sx, x, x, stat))) {
-      if (statP2X == 4) {
-        status = 4;
+      if (statP2X == SPCERR_BAD_SPEC) {
+        status = SPCERR_BAD_SPEC;
+      } else if (statP2X == SPXERR_BAD_SPEC_PARAMS) {
+        return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+          "Invalid spectral parameters: Frequency or wavelength is 0");
       } else {
-        return statP2X;
+        return wcserr_set(SPC_ERRMSG(statP2X));
       }
     }
   }
@@ -618,6 +679,9 @@ int spcs2x(
     *xp /= spc->w[2];
   }
 
+  if (status) {
+    wcserr_set(SPC_ERRMSG(status));
+  }
   return status;
 }
 
@@ -634,6 +698,26 @@ int spctyp(
   int  *restreq)
 
 {
+  return spctype(
+    ctypei, stype, scode, sname, units, ptype, xtype, restreq, NULL);
+}
+
+/* : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : :  */
+
+int spctype(
+  const char ctypei[9],
+  char stype[],
+  char scode[],
+  char sname[],
+  char units[],
+  char *ptype,
+  char *xtype,
+  int  *restreq,
+  struct wcserr **err)
+
+{
+  static const char *function = "spctype";
+
   char ctype[9], ptype_t, sname_t[32], units_t[8], xtype_t;
   int  restreq_t = 0;
 
@@ -690,7 +774,8 @@ int spctyp(
     strcpy(units_t, "");
     ptype_t = 'V';
   } else {
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Unknown spectral type '%s'", ctype);
   }
 
 
@@ -698,13 +783,15 @@ int spctyp(
   if ((xtype_t = ctype[5]) == ' ') {
     /* The algorithm code must be completely blank. */
     if (strcmp(ctype+4, "    ") != 0) {
-      return 2;
+      return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+        "Invalid spectral algorithm '%s'", ctype+4);
     }
 
     xtype_t = ptype_t;
 
   } else if (ctype[4] != '-') {
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Invalid spectral type '%s'", ctype);
 
   } else if (strcmp(ctype+5, "LOG") == 0 || strcmp(ctype+5, "TAB") == 0) {
     /* Logarithmic or tabular axis, not linear in any spectral type. */
@@ -712,7 +799,8 @@ int spctyp(
   } else if (xtype_t == 'G') {
     /* Validate the algorithm code. */
     if (ctype[6] != 'R') {
-      return 2;
+      return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+        "Invalid spectral algorithm '%s'", xtype_t);
     }
 
     /* Grism coordinates... */
@@ -723,16 +811,19 @@ int spctyp(
       /* ...in air. */
       xtype_t = 'a';
     } else {
-      return 2;
+      return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+        "Invalid spectral algorithm '%s'", xtype_t);
     }
 
   } else if (ctype[6] != '2') {
     /* Algorithm code has invalid syntax. */
-    return 2;
-
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Invalid spectral algorithm syntax '%s'", xtype_t);
   } else if (ctype[7] != ptype_t && ctype[7] != '?') {
     /* The P-, and S-type variables are inconsistent. */
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "In spectral type '%s', P- and S-type variables are inconsistent",
+      ctype);
 
   } else if (ctype[7] == ctype[5]) {
     /* Degenerate algorithm code. */
@@ -751,7 +842,8 @@ int spctyp(
     }
   } else if (strchr("LT", (int)xtype_t) == 0) {
     /* Invalid X-type variable code. */
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "In spectral type '%s', invalid X-type variable code", ctype);
   }
 
 
@@ -785,30 +877,62 @@ int spcspx(
   double *dXdS)
 
 {
+  return spcspxe(ctypeS, crvalS, restfrq, restwav, ptype, xtype, restreq,
+                 crvalX, dXdS, 0x0);
+}
+
+/* : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : :  */
+
+int spcspxe(
+  const char ctypeS[9],
+  double crvalS,
+  double restfrq,
+  double restwav,
+  char *ptype,
+  char *xtype,
+  int *restreq,
+  double *crvalX,
+  double *dXdS,
+  struct wcserr **err)
+
+{
+  static const char *function = "spcspxe";
+
   char scode[4], stype[5], type[8];
   int  status;
   double dPdS, dXdP;
   struct spxprm spx;
 
+
   /* Analyse the spectral axis code. */
-  if (spctyp(ctypeS, stype, scode, 0x0, 0x0, ptype, xtype, restreq)) {
-    return 2;
+  if ((status = spctype(ctypeS, stype, scode, 0x0, 0x0, ptype, xtype, restreq,
+                        err))) {
+    return status;
   }
 
   if (strstr("LT", xtype)) {
     /* Can't handle logarithmic or tabular coordinates. */
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Can't handle logarithmic or tabular coordinates");
   }
 
   /* Do we have rest frequency and/or wavelength as required? */
   if ((*restreq)%3 && restfrq == 0.0 && restwav == 0.0) {
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Missing required rest frequency or wavelength");
   }
 
   /* Compute all spectral parameters and their derivatives. */
   strcpy(type, stype);
+  spx.err = (err ? *err : 0x0);
   if ((status = specx(type, crvalS, restfrq, restwav, &spx))) {
-    return 2;
+    status = SPCERR_BAD_SPEC_PARAMS;
+    if (err) {
+      (*err)->status = status;
+    } else {
+      free(spx.err);
+    }
+    return status;
   }
 
 
@@ -925,24 +1049,48 @@ int spcxps(
   double *dSdX)
 
 {
+  return spcxpse(ctypeS, crvalX, restfrq, restwav, ptype, xtype, restreq,
+                 crvalS, dSdX, NULL);
+}
+
+/* : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : :  */
+
+int spcxpse(
+  const char ctypeS[9],
+  double crvalX,
+  double restfrq,
+  double restwav,
+  char *ptype,
+  char *xtype,
+  int *restreq,
+  double *crvalS,
+  double *dSdX,
+  struct wcserr **err)
+
+{
+  static const char *function = "spcxpse";
+
   char scode[4], stype[5], type[8];
   int  status;
   double dPdX, dSdP;
   struct spxprm spx;
 
   /* Analyse the spectral axis type. */
-  if (spctyp(ctypeS, stype, scode, 0x0, 0x0, ptype, xtype, restreq)) {
-    return 2;
+  if ((status = spctype(ctypeS, stype, scode, 0x0, 0x0, ptype, xtype, restreq,
+                        err))) {
+    return status;
   }
 
   if (strstr("LT", xtype)) {
     /* Can't handle logarithmic or tabular coordinates. */
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Can't handle logarithmic or tabular coordinates");
   }
 
   /* Do we have rest frequency and/or wavelength as required? */
   if ((*restreq)%3 && restfrq == 0.0 && restwav == 0.0) {
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Missing required rest frequency or wavelength");
   }
 
   /* Compute all spectral parameters and their derivatives. */
@@ -956,8 +1104,15 @@ int spcxps(
     strcpy(type, "VELO");
   }
 
-  if ((status = specx(type, crvalX, restfrq, restwav, &spx))) {
-    return 2;
+  spx.err = (err ? *err : 0x0);
+  if (specx(type, crvalX, restfrq, restwav, &spx)) {
+    status = SPCERR_BAD_SPEC_PARAMS;
+    if (err) {
+      (*err)->status = status;
+    } else {
+      free(spx.err);
+    }
+    return status;
   }
 
 
@@ -1068,12 +1223,32 @@ int spctrn(
   double *cdeltS2)
 
 {
+  return spctrne(ctypeS1, crvalS1, cdeltS1, restfrq, restwav,
+                 ctypeS2, crvalS2, cdeltS2, NULL);
+}
+
+/* : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : : :  */
+
+int spctrne(
+  const char ctypeS1[9],
+  double crvalS1,
+  double cdeltS1,
+  double restfrq,
+  double restwav,
+  char   ctypeS2[9],
+  double *crvalS2,
+  double *cdeltS2,
+  struct wcserr **err)
+
+{
+  static const char *function = "spctrne";
+
   char *cp, ptype1, ptype2, xtype1, xtype2;
   int  restreq, status;
   double crvalX, dS2dX, dXdS1;
 
-  if ((status = spcspx(ctypeS1, crvalS1, restfrq, restwav, &ptype1, &xtype1,
-                       &restreq, &crvalX, &dXdS1))) {
+  if ((status = spcspxe(ctypeS1, crvalS1, restfrq, restwav, &ptype1, &xtype1,
+                        &restreq, &crvalX, &dXdS1, err))) {
     return status;
   }
 
@@ -1094,14 +1269,15 @@ int spctrn(
     }
   }
 
-  if ((status = spcxps(ctypeS2, crvalX, restfrq, restwav, &ptype2, &xtype2,
-                       &restreq, crvalS2, &dS2dX))) {
+  if ((status = spcxpse(ctypeS2, crvalX, restfrq, restwav, &ptype2, &xtype2,
+                        &restreq, crvalS2, &dS2dX, err))) {
     return status;
   }
 
   /* Are the X-types compatible? */
   if (xtype2 != xtype1) {
-    return 2;
+    return wcserr_set(WCSERR_SET(SPCERR_BAD_SPEC_PARAMS),
+      "Incompatible X-types '%c' and '%c'", xtype1, xtype2);
   }
 
   if (ctypeS2[7] == '?') {
@@ -1142,7 +1318,7 @@ int spcaips(
       strncmp(ctype, "VELO", 4) == 0 ||
       strncmp(ctype, "FELO", 4) == 0) {
     /* Look for the Doppler frame. */
-    if ((fcode = ctype+4)) {
+    if (*(fcode = ctype+4)) {
       if (strcmp(fcode, "-LSR") == 0) {
         strcpy(specsys, "LSRK");
       } else if (strcmp(fcode, "-HEL") == 0) {
