@@ -1,6 +1,6 @@
 /*============================================================================
 
-  WCSLIB 4.7 - an implementation of the FITS WCS standard.
+  WCSLIB 4.8 - an implementation of the FITS WCS standard.
   Copyright (C) 1995-2011, Mark Calabretta
 
   This file is part of WCSLIB.
@@ -28,10 +28,10 @@
 
   Author: Mark Calabretta, Australia Telescope National Facility
   http://www.atnf.csiro.au/~mcalabre/index.html
-  $Id: spx.h,v 4.7 2011/02/07 07:03:42 cal103 Exp $
+  $Id: spx.h,v 4.8.1.1 2011/08/15 08:07:06 cal103 Exp cal103 $
 *=============================================================================
 *
-* WCSLIB 4.7 - C routines that implement the spectral coordinate systems
+* WCSLIB 4.8 - C routines that implement the spectral coordinate systems
 * recognized by the FITS World Coordinate System (WCS) standard.  Refer to
 *
 *   "Representations of world coordinates in FITS",
@@ -128,7 +128,9 @@
 *                       The type of spectral variable given by spec, FREQ,
 *                       AFRQ, ENER, WAVN, VRAD, WAVE, VOPT, ZOPT, AWAV, VELO,
 *                       or BETA (case sensitive).
+*
 *   spec      double    The spectral variable given, in SI units.
+*
 *   restfrq,
 *   restwav   double    Rest frequency [Hz] or rest wavelength in vacuo [m],
 *                       only one of which need be given.  The other should be
@@ -151,6 +153,8 @@
 *                         2: Invalid spectral parameters.
 *                         3: Invalid spectral variable.
 *
+*                       For returns > 1, a detailed error message is set in
+*                       spxprm::err if enabled, see wcserr_enable().
 *
 * freqafrq(), afrqfreq(), freqener(), enerfreq(), freqwavn(), wavnfreq(),
 * freqwave(), wavefreq(), freqawav(), awavfreq(), waveawav(), awavwave(),
@@ -165,14 +169,18 @@
 *
 * Given:
 *   param     double    Ignored.
+*
 *   nspec     int       Vector length.
+*
 *   instep,
 *   outstep   int       Vector strides.
+*
 *   inspec    const double[]
 *                       Input spectral variables, in SI units.
 *
 * Returned:
 *   outspec   double[]  Output spectral variables, in SI units.
+*
 *   stat      int[]     Status return value for each vector element:
 *                         0: Success.
 *                         1: Invalid value of inspec.
@@ -196,14 +204,18 @@
 *
 * Given:
 *   param     double    Rest frequency [Hz].
+*
 *   nspec     int       Vector length.
+*
 *   instep,
 *   outstep   int       Vector strides.
+*
 *   inspec    const double[]
 *                       Input spectral variables, in SI units.
 *
 * Returned:
 *   outspec   double[]  Output spectral variables, in SI units.
+*
 *   stat      int[]     Status return value for each vector element:
 *                         0: Success.
 *                         1: Invalid value of inspec.
@@ -227,14 +239,18 @@
 *
 * Given:
 *   param     double    Rest wavelength in vacuo [m].
+*
 *   nspec     int       Vector length.
+*
 *   instep,
 *   outstep   int       Vector strides.
+*
 *   inspec    const double[]
 *                       Input spectral variables, in SI units.
 *
 * Returned:
 *   outspec   double[]  Output spectral variables, in SI units.
+*
 *   stat      int[]     Status return value for each vector element:
 *                         0: Success.
 *                         1: Invalid value of inspec.
@@ -389,6 +405,12 @@
 *   double dbetavelo
 *     (Returned) ... vice versa [s/m] (constant, = 1/c, always available).
 *
+*   struct wcserr *err
+*     (Returned) If enabled, when an error status is returned this struct
+*     contains detailed information about the error, see wcserr_enable().
+*
+*   void *padding
+*     (An unused variable inserted for alignment purposes only.)
 *
 * Global variable: const char *spx_errmsg[] - Status return messages
 * ------------------------------------------------------------------
@@ -403,45 +425,62 @@
 extern "C" {
 #endif
 
+#include "wcserr.h"
 
 extern const char *spx_errmsg[];
 
+enum spx_errmsg {
+  SPXERR_SUCCESS          = 0,	/* Success. */
+  SPXERR_NULL_POINTER     = 1,	/* Null spxprm pointer passed. */
+  SPXERR_BAD_SPEC_PARAMS  = 2,	/* Invalid spectral parameters. */
+  SPXERR_BAD_SPEC_VAR     = 3,	/* Invalid spectral variable. */
+  SPXERR_BAD_INSPEC_COORD = 4 	/* One or more of the inspec coordinates were
+				   invalid. */
+};
 
 struct spxprm {
-  double restfrq, restwav;      /* Rest frequency [Hz] and wavelength [m].  */
+  double restfrq, restwav;	/* Rest frequency [Hz] and wavelength [m].  */
 
-  int wavetype, velotype;       /* True if wave/velocity types have been    */
-                                /* computed; types are defined below.       */
+  int wavetype, velotype;	/* True if wave/velocity types have been    */
+				/* computed; types are defined below.       */
 
   /* Spectral variables computed by specx().                                */
   /*------------------------------------------------------------------------*/
-  double freq,                  /* wavetype: Frequency [Hz].                */
-         afrq,                  /* wavetype: Angular frequency [rad/s].     */
-         ener,                  /* wavetype: Photon energy [J].             */
-         wavn,                  /* wavetype: Wave number [/m].              */
-         vrad,                  /* velotype: Radio velocity [m/s].          */
-         wave,                  /* wavetype: Vacuum wavelength [m].         */
-         vopt,                  /* velotype: Optical velocity [m/s].        */
-         zopt,                  /* velotype: Redshift.                      */
-         awav,                  /* wavetype: Air wavelength [m].            */
-         velo,                  /* velotype: Relativistic velocity [m/s].   */
-         beta;                  /* velotype: Relativistic beta.             */
+  double freq,			/* wavetype: Frequency [Hz].                */
+         afrq,			/* wavetype: Angular frequency [rad/s].     */
+         ener,			/* wavetype: Photon energy [J].             */
+         wavn,			/* wavetype: Wave number [/m].              */
+         vrad,			/* velotype: Radio velocity [m/s].          */
+         wave,			/* wavetype: Vacuum wavelength [m].         */
+         vopt,			/* velotype: Optical velocity [m/s].        */
+         zopt,			/* velotype: Redshift.                      */
+         awav,			/* wavetype: Air wavelength [m].            */
+         velo,			/* velotype: Relativistic velocity [m/s].   */
+         beta;			/* velotype: Relativistic beta.             */
 
   /* Derivatives of spectral variables computed by specx().                 */
   /*------------------------------------------------------------------------*/
-  double dfreqafrq, dafrqfreq,  /* Constant, always available.              */
-         dfreqener, denerfreq,  /* Constant, always available.              */
-         dfreqwavn, dwavnfreq,  /* Constant, always available.              */
-         dfreqvrad, dvradfreq,  /* wavetype && velotype.                    */
-         dfreqwave, dwavefreq,  /* wavetype.                                */
-         dfreqawav, dawavfreq,  /* wavetype.                                */
-         dfreqvelo, dvelofreq,  /* wavetype && velotype.                    */
-         dwavevopt, dvoptwave,  /* wavetype && velotype.                    */
-         dwavezopt, dzoptwave,  /* wavetype && velotype.                    */
-         dwaveawav, dawavwave,  /* wavetype.                                */
-         dwavevelo, dvelowave,  /* wavetype && velotype.                    */
-         dawavvelo, dveloawav,  /* wavetype && velotype.                    */
-         dvelobeta, dbetavelo;  /* Constant, always available.              */
+  double dfreqafrq, dafrqfreq,	/* Constant, always available.              */
+         dfreqener, denerfreq,	/* Constant, always available.              */
+         dfreqwavn, dwavnfreq,	/* Constant, always available.              */
+         dfreqvrad, dvradfreq,	/* wavetype && velotype.                    */
+         dfreqwave, dwavefreq,	/* wavetype.                                */
+         dfreqawav, dawavfreq,	/* wavetype.                                */
+         dfreqvelo, dvelofreq,	/* wavetype && velotype.                    */
+         dwavevopt, dvoptwave,	/* wavetype && velotype.                    */
+         dwavezopt, dzoptwave,	/* wavetype && velotype.                    */
+         dwaveawav, dawavwave,	/* wavetype.                                */
+         dwavevelo, dvelowave,	/* wavetype && velotype.                    */
+         dawavvelo, dveloawav,	/* wavetype && velotype.                    */
+         dvelobeta, dbetavelo;	/* Constant, always available.              */
+
+  /* Error handling                                                         */
+  /*------------------------------------------------------------------------*/
+  struct wcserr *err;
+
+  /* Private                                                                */
+  /*------------------------------------------------------------------------*/
+  void   *padding;		/* (Dummy inserted for alignment purposes.) */
 };
 
 /* Size of the spxprm struct in int units, used by the Fortran wrappers. */
@@ -454,7 +493,7 @@ int specx(const char *type, double spec, double restfrq, double restwav,
 
 /* For use in declaring function prototypes, e.g. in spcprm. */
 #define SPX_ARGS double param, int nspec, int instep, int outstep, \
-                 const double inspec[], double outspec[], int stat[]
+    const double inspec[], double outspec[], int stat[]
 
 int freqafrq(SPX_ARGS);
 int afrqfreq(SPX_ARGS);
