@@ -1,6 +1,6 @@
 /*============================================================================
 
-  WCSLIB 4.7 - an implementation of the FITS WCS standard.
+  WCSLIB 4.8 - an implementation of the FITS WCS standard.
   Copyright (C) 1995-2011, Mark Calabretta
 
   This file is part of WCSLIB.
@@ -28,10 +28,10 @@
 
   Author: Mark Calabretta, Australia Telescope National Facility
   http://www.atnf.csiro.au/~mcalabre/index.html
-  $Id: tab.h,v 4.7 2011/02/07 07:03:42 cal103 Exp $
+  $Id: tab.h,v 4.8.1.1 2011/08/15 08:07:06 cal103 Exp cal103 $
 *=============================================================================
 *
-* WCSLIB 4.7 - C routines that implement tabular coordinate systems as
+* WCSLIB 4.8 - C routines that implement tabular coordinate systems as
 * defined by the FITS World Coordinate System (WCS) standard.  Refer to
 *
 *   "Representations of world coordinates in FITS",
@@ -97,6 +97,7 @@
 *                       saves having to initalize these pointers to zero.)
 *
 *   M         int       The number of tabular coordinate axes.
+*
 *   K         const int[]
 *                       Vector of length M whose elements (K_1, K_2,... K_M)
 *                       record the lengths of the axes of the coordinate array
@@ -128,6 +129,9 @@
 *                         2: Memory allocation failed.
 *                         3: Invalid tabular parameters.
 *
+*                       For returns > 1, a detailed error message is set in
+*                       tabprm::err if enabled, see wcserr_enable().
+*
 *
 * tabmem() - Acquire tabular memory
 * ---------------------------------
@@ -142,6 +146,10 @@
 *             int       Status return value:
 *                         0: Success.
 *                         1: Null tabprm pointer passed.
+*                         2: Memory allocation failed.
+*
+*                       For returns > 1, a detailed error message is set in
+*                       tabprm::err if enabled, see wcserr_enable().
 *
 *
 * tabcpy() - Copy routine for the tabprm struct
@@ -176,6 +184,10 @@
 *                         1: Null tabprm pointer passed.
 *                         2: Memory allocation failed.
 *
+*                       For returns > 1, a detailed error message is set in
+*                       tabprm::err (associated with tabdst) if enabled, see
+*                       wcserr_enable().
+*
 *
 * tabfree() - Destructor for the tabprm struct
 * --------------------------------------------
@@ -198,7 +210,8 @@
 *
 * tabprt() - Print routine for the tabprm struct
 * ----------------------------------------------
-* tabprt() prints the contents of a tabprm struct.
+* tabprt() prints the contents of a tabprm struct using wcsprintf().  Mainly
+* intended for diagnostic purposes.
 *
 * Given:
 *   tab       const struct tabprm*
@@ -229,6 +242,9 @@
 *                         1: Null tabprm pointer passed.
 *                         3: Invalid tabular parameters.
 *
+*                       For returns > 1, a detailed error message is set in
+*                       tabprm::err if enabled, see wcserr_enable().
+*
 *
 * tabx2s() - Pixel-to-world transformation
 * ----------------------------------------
@@ -243,12 +259,14 @@
 *   ncoord,
 *   nelem     int       The number of coordinates, each of vector length
 *                       nelem.
+*
 *   x         const double[ncoord][nelem]
 *                       Array of intermediate world coordinates, SI units.
 *
 * Returned:
 *   world     double[ncoord][nelem]
 *                       Array of world coordinates, in SI units.
+*
 *   stat      int[ncoord]
 *                       Status return value status for each coordinate:
 *                         0: Success.
@@ -262,6 +280,9 @@
 *                         4: One or more of the x coordinates were invalid,
 *                            as indicated by the stat vector.
 *
+*                       For returns > 1, a detailed error message is set in
+*                       tabprm::err if enabled, see wcserr_enable().
+*
 *
 * tabs2x() - World-to-pixel transformation
 * ----------------------------------------
@@ -294,6 +315,9 @@
 *                         5: One or more of the world coordinates were
 *                            invalid, as indicated by the stat vector.
 *
+*                       For returns > 1, a detailed error message is set in
+*                       tabprm::err if enabled, see wcserr_enable().
+*
 *
 * tabprm struct - Tabular transformation parameters
 * -------------------------------------------------
@@ -399,7 +423,7 @@
 *     whose elements indicate whether the corresponding indexing vector is
 *     monotonic increasing (+1), or decreasing (-1).
 *
-*   double *p0
+*   int *p0
 *     (Returned) Pointer to the first element of a vector of length tabprm::M
 *     of interpolated indices into the coordinate array such that Upsilon_m,
 *     as defined in Paper III, is equal to (p0[m] + 1) + tabprm::delta[m].
@@ -420,6 +444,10 @@
 *     compressed K_1 dimension, then the maximum.  This array is used by the
 *     inverse table lookup function, tabs2x(), to speed up table searches.
 *
+*   struct wcserr *err
+*     (Returned) If enabled, when an error status is returned this struct
+*     contains detailed information about the error, see wcserr_enable().
+*
 *   int m_flag
 *     (For internal use only.)
 *   int m_M
@@ -451,6 +479,8 @@
 #ifndef WCSLIB_TAB
 #define WCSLIB_TAB
 
+#include "wcserr.h"
+
 #ifdef __cplusplus
 extern "C" {
 #endif
@@ -458,44 +488,60 @@ extern "C" {
 
 extern const char *tab_errmsg[];
 
+enum tab_errmsg_enum {
+  TABERR_SUCCESS      = 0,	/* Success. */
+  TABERR_NULL_POINTER = 1,	/* Null tabprm pointer passed. */
+  TABERR_MEMORY       = 2,	/* Memory allocation failed. */
+  TABERR_BAD_PARAMS   = 3,	/* Invalid tabular parameters. */
+  TABERR_BAD_X        = 4,	/* One or more of the x coordinates were
+				   invalid. */
+  TABERR_BAD_WORLD    = 5	/* One or more of the world coordinates were
+				   invalid. */
+};
 
 struct tabprm {
   /* Initialization flag (see the prologue above).                          */
   /*------------------------------------------------------------------------*/
-  int    flag;                  /* Set to zero to force initialization.     */
+  int    flag;			/* Set to zero to force initialization.     */
 
   /* Parameters to be provided (see the prologue above).                    */
   /*------------------------------------------------------------------------*/
-  int    M;                     /* Number of tabular coordinate axes.       */
-  int    *K;                    /* Vector of length M whose elements        */
-                                /* (K_1, K_2,... K_M) record the lengths of */
-                                /* the axes of the coordinate array and of  */
-                                /* each indexing vector.                    */
-  int    *map;                  /* Vector of length M usually such that     */
-                                /* map[m-1] == i-1 for coordinate array     */
-                                /* axis m and image axis i (see above).     */
-  double *crval;                /* Vector of length M containing the index  */
-                                /* value for the reference pixel for each   */
-                                /* of the tabular coordinate axes.          */
-  double **index;               /* Vector of pointers to M indexing vectors */
-                                /* of lengths (K_1, K_2,... K_M).           */
-  double *coord;                /* (1+M)-dimensional tabular coordinate     */
-                                /* array (see above).                       */
+  int    M;			/* Number of tabular coordinate axes.       */
+  int    *K;			/* Vector of length M whose elements        */
+				/* (K_1, K_2,... K_M) record the lengths of */
+				/* the axes of the coordinate array and of  */
+				/* each indexing vector.                    */
+  int    *map;			/* Vector of length M usually such that     */
+				/* map[m-1] == i-1 for coordinate array     */
+				/* axis m and image axis i (see above).     */
+  double *crval;		/* Vector of length M containing the index  */
+				/* value for the reference pixel for each   */
+				/* of the tabular coordinate axes.          */
+  double **index;		/* Vector of pointers to M indexing vectors */
+				/* of lengths (K_1, K_2,... K_M).           */
+  double *coord;		/* (1+M)-dimensional tabular coordinate     */
+				/* array (see above).                       */
 
   /* Information derived from the parameters supplied.                      */
   /*------------------------------------------------------------------------*/
-  int    nc;                    /* Number of coordinate vectors (of length  */
-                                /* M) in the coordinate array.              */
-  int    padding;               /* (Dummy inserted for alignment purposes.) */
-  int    *sense;                /* Vector of M flags that indicate whether  */
-                                /* the Mth indexing vector is monotonic     */
-                                /* increasing, or else decreasing.          */
-  int    *p0;                   /* Vector of M indices.                     */
-  double *delta;                /* Vector of M increments.                  */
-  double *extrema;              /* (1+M)-dimensional array of coordinate    */
-                                /* extrema.                                 */
+  int    nc;			/* Number of coordinate vectors (of length  */
+				/* M) in the coordinate array.              */
+  int    padding;		/* (Dummy inserted for alignment purposes.) */
+  int    *sense;		/* Vector of M flags that indicate whether  */
+				/* the Mth indexing vector is monotonic     */
+				/* increasing, or else decreasing.          */
+  int    *p0;			/* Vector of M indices.                     */
+  double *delta;		/* Vector of M increments.                  */
+  double *extrema;		/* (1+M)-dimensional array of coordinate    */
+				/* extrema.                                 */
 
-  int    m_flag, m_M, m_N;      /* The remainder are for memory management. */
+  /* Error handling                                                         */
+  /*------------------------------------------------------------------------*/
+  struct wcserr *err;
+
+  /* Private - the remainder are for memory management.                     */
+  /*------------------------------------------------------------------------*/
+  int    m_flag, m_M, m_N;
   int    set_M;
   int    *m_K, *m_map;
   double *m_crval, **m_index, **m_indxs, *m_coord;
